from unittest import TestCase

import click
from click.testing import CliRunner, Result
from logging import Logger, getLogger
from inspect import stack

from pandas import DataFrame
from pandas.testing import assert_frame_equal

from com_enovation.helper.click.df_modify import df_compress


@click.group(chain=True)
@click.pass_context
def test_click(ctx_context):

    # Ensure that ctx_context.obj exists and is a dict (in case `cli()` is called
    # by means other than the `if` block below)
    # This is effectively the context, that is shared across commands
    ctx_context.ensure_object(dict)

    # We set the test data for alias "jsg"
    # Same data set as in:
    # - file "tests/com/enovation/helper/test_pandas_dataframe_sampler/test_01_dataframe_sampler.py"
    # - test function "test_06_full_example_reordered_keep_last"
    ctx_context.obj["jsg"] = DataFrame(
        columns=[
            "key 1", "key 2", "order 1", "order 2", "val 1", "val 2"
        ],
        data=[
            # Universe 1
            ["k1a", "k2a", "o1b", "o2a", "v1b", "v2a"],   # 3 - in
            ["k1a", "k2a", "o1b", "o2c", "v1a", "v2a"],  # 5 - out
            ["k1a", "k2a", "o1a", "o2a", "v1a", "v2a"],   # 1 - in
            ["k1a", "k2a", "o1c", "o2b", "v1a", "v2a"],   # 6 - in
            ["k1a", "k2a", "o1c", "o2a", "v1a", "v2a"],  # 5 - out
            ["k1a", "k2a", "o1a", "o2b", "v1a", "v2a"],  # 2 - out
            ["k1a", "k2a", "o1b", "o2b", "v1a", "v2a"],   # 4 - in
        ]
    )


@click.command('test-check-click')
@click.pass_context
def test_check_click(ctx_context):
    df_expected: DataFrame = DataFrame(
        columns=[
            "key 1", "key 2", "order 1", "order 2", "val 1", "val 2", "expected result"
        ],
        data=[
            # Universe 1
            ["k1a", "k2a", "o1b", "o2a", "v1b", "v2a", 2],   # 3 - in
            ["k1a", "k2a", "o1b", "o2c", "v1a", "v2a", -1],  # 5 - out
            ["k1a", "k2a", "o1a", "o2a", "v1a", "v2a", 1],   # 1 - in
            ["k1a", "k2a", "o1c", "o2b", "v1a", "v2a", 4],   # 6 - in
            ["k1a", "k2a", "o1c", "o2a", "v1a", "v2a", -1],  # 5 - out
            ["k1a", "k2a", "o1a", "o2b", "v1a", "v2a", -1],  # 2 - out
            ["k1a", "k2a", "o1b", "o2b", "v1a", "v2a", 3],   # 4 - in
        ]
    )

    assert_frame_equal(
        left=df_expected[
            df_expected["expected result"] > 0
            ].sort_values(by="expected result")[ctx_context.obj["jsg"].columns].reset_index(drop=True),
        right=ctx_context.obj["jsg"].reset_index(drop=True),
        obj=f"Call 1",
        atol=0.00001,
        check_like=True  # To exclude index, and not to care about column ordering
    )


# We add the click command we want to test
# noinspection PyTypeChecker
test_click.add_command(df_compress)
test_click.add_command(test_check_click)


class TestFunctionDropDuplicates(TestCase):

    _logger: Logger = getLogger(__name__)

    def test_01_full_example_reordered_keep_last(self):
        self._logger.debug(f"Function '{stack()[0].filename} - {stack()[0].function}' is called.")

        """
        Note: When calling function CliRunner.invoke, we add a tag "# noinspection PyTypeChecker" not to raise a
        warning due to function "invoke" expecting function that we call to be a "BaseCommand", which does not seem to
        be the case...
        :return:
        """
        self._logger.debug(f"Function '{stack()[0].filename} - {stack()[0].function}' is called")

        the_runner = CliRunner()

        # The click line is in 3 steps:
        # - First, the group "test_click", to instantiate the dataframe into the context
        # - Second, call the command drop_duplicates
        # - Third, call a test command to reconcile expected results
        the_result: Result = the_runner.invoke(
            cli=test_click,
            args=[
                'df-compress',
                '-k', '["key 1", "key 2"]',
                '-r', '["order 1", "order 2"]',
                '-v', '["val 1", "val 2"]',
                '--keep-last',
                'jsg',
                'test-check-click'
            ],
            catch_exceptions=False
        )
        self.assertEqual(
            0,
            the_result.exit_code,
            f"Unexpected exit_code, with stdout: {the_result.stdout}"
        )

        self._logger.debug(f"Function '{stack()[0].filename} - {stack()[0].function}' is returning.")



import os
import unittest
from pathlib import Path
from logging import Logger, getLogger
from inspect import stack
from pandas import DataFrame, read_excel

from com_enovation.toolbox.data_handler.data_modifier.data_modification_feature_factory \
    import DataModificationFeatureFactory


class TestDataModificationFeatureFactory_Filter(unittest.TestCase):
    _logger: Logger = getLogger(__name__)

    def test_filter(self):
        self._logger.debug(f"Function '{stack()[0].filename} - {stack()[0].function}' is called")

        # We load the data extract
        df_data_extract: DataFrame = read_excel(
            Path(os.path.join(os.path.dirname(__file__),
                              '01.data_extract.xlsx'))
        )

        df_the_modified_dataframe: DataFrame = DataModificationFeatureFactory(). \
            get_data_modification_feature__filter(
            str_expressions=f"""
                StatusA+StatusB=={[
                "OpenOpen",
                "OpenInactive",
                "InactiveOpen",
                "InactiveInactive",
                "ClosedClosed"
            ]}
                """,
            b_copy=True
        )(df_data_extract)

        self.assertEqual(
            first=df_the_modified_dataframe.to_string(),
            second="""    StatusA   StatusB
0      Open      Open
1      Open  Inactive
3  Inactive      Open
4  Inactive  Inactive
8    Closed    Closed""",
            msg=""
        )
        self._logger.debug(f"Function '{stack()[0].filename} - {stack()[0].function}' is returning")



import os
import signal
import unittest
from pathlib import Path
from time import sleep

from click.testing import CliRunner, Result
from logging import Logger, getLogger
from inspect import stack

from pandas.testing import assert_frame_equal

from com_enovation.toolbox.predictability.dp_date_predictability.dp_computer import PredictabilityBean
from com_enovation.toolbox.predictability.dp_date_predictability.dp_persister import DatePredictabilityPersister
from com_enovation.enov import enov


class TestDatePredictability(unittest.TestCase):

    _logger: Logger = getLogger(__name__)

    def test_date_predictability(self):
        """
        Note: When calling function 'CliRunner.invoke', we add a tag "# noinspection PyTypeChecker" not to raise a
        warning due to function "invoke" expecting function that we call to be a "BaseCommand", which does not seem to
        be the case...
        :return:
        """
        self._logger.debug(f"Function '{stack()[0].filename} - {stack()[0].function}' is called")

        # First, we cleanse the output file from previous run
        if Path(os.path.join(os.path.dirname(__file__), '03_actual_results.xlsx')).exists():
            os.remove(os.path.join(os.path.dirname(__file__), '03_actual_results.xlsx'))
        self.assertFalse(
            Path(os.path.join(os.path.dirname(__file__), '03_actual_results.xlsx')).exists(),
            f"We still have a file '{os.path.join(os.path.dirname(__file__), '03_actual_results.xlsx')}', that will "
            f"generate issue while executing the test..."
        )

        the_runner = CliRunner()

        # ###############################################################
        # Step 1: we compute date predictability, and we persist the bean
        # noinspection PyTypeChecker
        the_result: Result = the_runner.invoke(
            cli=enov,
            args=[
                # 'enov-warm-up',
                '--verbose',

                # We load the Excel file
                'df-load-xls',
                '-c', '["id","timestamp","kick_off_date"]',
                os.path.join(os.path.dirname(__file__), '01_input_file.xlsx'),
                'df_raw',

                # We rename the columns
                'df-rename-columns',
                '-o', 'df_renamed',
                'df_raw',
                '{"id":"key","timestamp":"date","kick_off_date":"measure"}',

                # We cleanse null values
                'df-cleanse-null',
                '-o', 'df_cleansed',
                '-c', '["measure"]',
                'df_renamed',
                
                # # We compute date predictability
                'dp-compute',
                '--resample',
                'df_cleansed',
                'df_predictability',

                # # We persist the predictability bean
                'dp-persist',
                'df_predictability',
                os.path.join(os.path.dirname(__file__), '03_actual_results.xlsx'),
            ],
            catch_exceptions=True
        )
        self.assertEqual(
            0,
            the_result.exit_code,
            f"Unexpected exit_code, with stdout: {the_result.stdout}"
        )
        self.assertTrue(
            os.path.isfile(
                os.path.join(os.path.dirname(__file__), '03_actual_results.xlsx')
            ),
            f"File '03_actual_results.xlsx' was not persisted..., with stdout: {the_result.stdout}"
        )

        # ###############################################################
        # Step 2: we reconcile the persisted bean
        obj_the_expected_results: PredictabilityBean = DatePredictabilityPersister().load(
            p_file_path=Path(os.path.join(os.path.dirname(__file__), '02_expected_results.xlsx')),
        )
        obj_the_actual_results: PredictabilityBean = DatePredictabilityPersister().load(
            p_file_path=Path(os.path.join(os.path.dirname(__file__), '03_actual_results.xlsx')),
        )
        assert_frame_equal(
            left=obj_the_expected_results.df_by_key,
            right=obj_the_actual_results._df_by_key,
            obj=f"The sheet {DatePredictabilityPersister.str_sheet_name__by_key} does not reconcile.",
            atol=0.00001,
            check_like=True  # To exclude index, and not to care about column ordering
        )
        assert_frame_equal(
            left=obj_the_expected_results.df_by_measure,
            right=obj_the_actual_results.df_by_measure,
            obj=f"The sheet {DatePredictabilityPersister.str_sheet_name__by_measure} does not reconcile.",
            atol=0.00001,
            check_like=True  # To exclude index, and not to care about column ordering
        )
        assert_frame_equal(
            left=obj_the_expected_results.df_historical,
            right=obj_the_actual_results.df_historical,
            obj=f"The sheet {DatePredictabilityPersister.str_sheet_name__historical} does not reconcile.",
            atol=0.00001,
            check_like=True  # To exclude index, and not to care about column ordering
        )

        # ###############################################################
        # 3. We launch the dp_grapher
        if hasattr(os, 'fork'):
            childpid = os.fork()

            if childpid == 0:
                # noinspection PyTypeChecker
                CliRunner().invoke(
                        cli=enov,
                        args=[
                            '--verbose',

                            # We load the predictability bean
                            'dp-load',
                            os.path.join(os.path.dirname(__file__), '03_actual_results.xlsx'),
                            'df_predictability',

                            # We graph
                            'dp-graph',
                            'df_predictability'
                        ],
                        catch_exceptions=True
                    )

            else:
                # in the parent
                sleep(10)
                os.kill(childpid, signal.SIGINT)

        else:
            self._logger.warning(f"You must be executing on Windows, as os.fork() could be executed...")

        self.assertEqual(
            0,
            the_result.exit_code,
            f"Unexpected exit_code, with stdout: {the_result.stdout}"
        )

        self._logger.debug(f"Function '{stack()[0].filename} - {stack()[0].function}' is returning")



import unittest
from inspect import stack
from logging import getLogger, Logger

from com_enovation.toolbox.excel_dashboard.cell_notation_helper import CellNotationHelper


class TestFunction__IsAddressXxx(unittest.TestCase):

    _logger: Logger = getLogger(__name__)

    def test_function__transform_column_into_range_or_cell(self):
        self._logger.debug(f"Function '{stack()[0].filename} - {stack()[0].function}' is called")

        # 1. Correct columns: pattern and value
        for k, v in {
            "A:B": "A5:B15",
            "A:A": "A5:A15",
            "D:BC": "D5:BC15",

            # With @
            "@A:B": "@A5:B15",
            "D:@BC": "D5:@BC15",

            # With $
            "$A:B": "$A5:B15",
            "D:$BC": "D5:$BC15",
        }.items():
            self.assertEqual(
                first=v,
                second=CellNotationHelper.transform_column_into_range_or_cell(
                    str_column=k,
                    int_first_row=5,
                    int_last_row=15,
                    b_check_correct=True
                ),
                msg=f"Error when process address '{k}'."
            )
            self.assertEqual(
                first=v,
                second=CellNotationHelper.transform_column_into_range_or_cell(
                    str_column=k,
                    int_first_row=5,
                    int_last_row=15,
                    b_check_correct=False
                ),
                msg=f"Error when process address '{k}'."
            )

        # 2. Correct pattern, but not correct columns
        for k, v in {
            "B:A": "B5:A15",
            "BC:D": "BC5:D15",

            # With @
            "@B:A": "@B5:A15",
            "BC:@D": "BC5:@D15",

            # With $
            "$B:A": "$B5:A15",
            "BC:$D": "BC5:$D15",
        }.items():
            # self.assertEqual(
            #     first=v,
            #     second=CellNotationHelper.transform_column_into_range_or_cell(
            #         str_column=k,
            #         int_first_row=5,
            #         int_last_row=15,
            #         b_check_correct=True
            #     ),
            #     msg=f"Error when process address '{k}'."
            # )
            self.assertEqual(
                first=v,
                second=CellNotationHelper.transform_column_into_range_or_cell(
                    str_column=k,
                    int_first_row=5,
                    int_last_row=15,
                    b_check_correct=False
                ),
                msg=f"Error when process address '{k}'."
            )

        # 3. Returning a cell, not a range
        for k, v in {
            "A:A": "A5",

            # With @
            "@A:@A": "@A5",

            # With $
            "$A:$A": "$A5"
        }.items():
            self.assertEqual(
                first=v,
                second=CellNotationHelper.transform_column_into_range_or_cell(
                    str_column=k,
                    int_first_row=5,
                    int_last_row=5,
                    b_check_correct=True
                ),
                msg=f"Error when process address '{k}'."
            )
            self.assertEqual(
                first=v,
                second=CellNotationHelper.transform_column_into_range_or_cell(
                    str_column=k,
                    int_first_row=5,
                    int_last_row=5,
                    b_check_correct=False
                ),
                msg=f"Error when process address '{k}'."
            )

        self._logger.debug(f"Function '{stack()[0].filename} - {stack()[0].function}' is returning")

    def test_function__transform_column_into_range__exceptions(self):
        self._logger.debug(f"Function '{stack()[0].filename} - {stack()[0].function}' is called")

        # 1. Crazy values
        for i_str_address in [
            "abc",
            " A3",
        ]:
            with self.assertRaisesRegex(Exception, f"The address '{i_str_address}' is not recognized as an address, "
                                                   f"and its pattern could not be determined."):
                CellNotationHelper.transform_column_into_range_or_cell(
                    str_column=i_str_address,
                    int_first_row=5,
                    int_last_row=15,
                    b_check_correct=True
                )
            with self.assertRaisesRegex(Exception, f"The address '{i_str_address}' is not recognized as an address, "
                                                   f"and its pattern could not be determined."):
                CellNotationHelper.transform_column_into_range_or_cell(
                    str_column=i_str_address,
                    int_first_row=5,
                    int_last_row=15,
                    b_check_correct=False
                )

        # 2. Not columns
        for i_str_address in [
            "A3",
            "A1:B2",
            "1:3"
        ]:
            with self.assertRaisesRegex(Exception,
                                        f"The parameter str_address '{i_str_address}' is not a column."):
                CellNotationHelper.transform_column_into_range_or_cell(
                    str_column=i_str_address,
                    int_first_row=5,
                    int_last_row=15,
                    b_check_correct=True
                )
            with self.assertRaisesRegex(Exception,
                                        f"The parameter str_address '{i_str_address}' is not a column."):
                CellNotationHelper.transform_column_into_range_or_cell(
                    str_column=i_str_address,
                    int_first_row=5,
                    int_last_row=15,
                    b_check_correct=False
                )

        # 3. Correct pattern, but not correct columns
        for k, v in {
            "B:A": "B5:A15",
            "BC:D": "BC5:D15",

            # With @
            "@B:A": "@B5:A15",
            "BC:@D": "BC5:@D15",

            # With $
            "$B:A": "$B5:A15",
            "BC:$D": "BC5:$D15",
        }.items():

            with self.assertRaisesRegex(Exception,
                                        f"The parameter str_address .* is not a correct column."):
                CellNotationHelper.transform_column_into_range_or_cell(
                    str_column=k,
                    int_first_row=5,
                    int_last_row=15,
                    b_check_correct=True
                )

        self._logger.debug(f"Function '{stack()[0].filename} - {stack()[0].function}' is returning")

    def test_function__transform_columns_intersect_rows_into_range_or_cell(self):
        self._logger.debug(f"Function '{stack()[0].filename} - {stack()[0].function}' is called")

        for i_test in [
            {
                "label": "TC 01 - Vanilla use case",
                "col": "A:B", "row": "1:2",  # "check": True,
                "expected": "A1:B2"
            },
            {
                "label": "TC 02 - Returning a cell, not a range",
                "col": "A:A", "row": "3:3",  # "check": True,
                "expected": "A3"
            },
            {
                "label": "TC 03 - With @",
                "col": "@A:B", "row": "1:2", "check": False,
                "expected": "@A1:B2"
            },
            {
                "label": "TC 04 - A range, not a cell, due to the @",
                "col": "@A:A", "row": "3:3", "check": False,
                "expected": "@A3:A3"
            },
            {
                "label": "TC 05 - With $",
                "col": "A:$B", "row": "$1:2",  # "check": True,
                "expected": "A$1:$B2"
            },
            {
                "label": "TC 06 - Returning a cell, not a range",
                "col": "$A:$A", "row": "3:3",  # "check": True,
                "expected": "$A3"
            },
            {
                "label": "TC 07 - With @ and $",
                "col": "@$A:$B", "row": "1:$2", "check": False,
                "expected": "@$A1:$B$2"
            },
            {
                "label": "TC 08 - A range, not a cell, due to the $",
                "col": "$A:A", "row": "3:3",  # "check": True,
                "expected": "$A3:A3"
            },
        ]:
            if i_test.get("check", True) is True:
                self.assertEqual(
                    first=i_test["expected"],
                    second=CellNotationHelper.transform_columns_intersect_rows_into_range_or_cell(
                        str_columns=i_test["col"],
                        str_rows=i_test["row"],
                        b_check_correct=True
                    ),
                    msg=f"Error when process address test with label '{i_test['label']}', b_check_correct is True."
                )
            if i_test.get("check", False) is False:
                self.assertEqual(
                    first=i_test["expected"],
                    second=CellNotationHelper.transform_columns_intersect_rows_into_range_or_cell(
                        str_columns=i_test["col"],
                        str_rows=i_test["row"],
                        b_check_correct=False
                    ),
                    msg=f"Error when process address test with label '{i_test['label']}', b_check_correct is False."
                )

        self._logger.debug(f"Function '{stack()[0].filename} - {stack()[0].function}' is returning")

    def test_function__transform_columns_intersect_rows_into_range__exceptions(self):
        self._logger.debug(f"Function '{stack()[0].filename} - {stack()[0].function}' is called")

        for i_test in [
            {
                "label": "TC 01 - totally wrong column",
                "col": "abc", "row": "1:2",  # "check": True,
                "exception": "The address 'abc' is not recognized as an address, and its pattern could not be "
                             "determined."
            },
            {
                "label": "TC 02 - totally wrong row",
                "col": "C:D", "row": "12",  # "check": True,
                "exception": "The address '12' is not recognized as an address, and its pattern could not be "
                             "determined."
            },
            {
                "label": "TC 03 - wrong column",
                "col": "D:C", "row": "1:2", "check": True,
                "exception": "The parameter str_columns 'D:C' is not a correct set of columns."
            },
            {
                "label": "TC 04 - wrong row",
                "col": "C:D", "row": "2:1", "check": True,
                "exception": "The parameter str_rows '2:1' is not a correct set of rows."
            },
        ]:
            if i_test.get("check", True) is True:
                with self.assertRaisesRegex(Exception, i_test["exception"]):
                    CellNotationHelper.transform_columns_intersect_rows_into_range_or_cell(
                        str_columns=i_test["col"],
                        str_rows=i_test["row"],
                        b_check_correct=True
                    )
            if i_test.get("check", False) is False:
                with self.assertRaisesRegex(Exception, i_test["exception"]):

                    CellNotationHelper.transform_columns_intersect_rows_into_range_or_cell(
                        str_columns=i_test["col"],
                        str_rows=i_test["row"],
                        b_check_correct=False
                    )

        self._logger.debug(f"Function '{stack()[0].filename} - {stack()[0].function}' is returning")

    def test_function__transform_address_into_relative_and_absolute_address(self):
        self._logger.debug(f"Function '{stack()[0].filename} - {stack()[0].function}' is called")

        for i_test in [
            {
                "label": "TC 01",
                "parameters": {
                    "str_address": "A1:B2",
                    "b_first_col_at": True, "b_first_col_doll": True,
                    "b_last_col_at": True, "b_last_col_doll": True,
                    "b_first_row_at": True, "b_first_row_doll": True,
                    "b_last_row_at": True, "b_last_row_doll": True,
                },
                "result": "@$A@$1:@$B@$2"
                # "exception": "The address 'abc' is not recognized as an address, and its pattern could not be "
                #             "determined."
            },
            {
                "label": "TC 02",
                "parameters": {
                    "str_address": "A1:B2",
                    "b_first_col_at": True, "b_first_col_doll": False,
                    "b_last_col_at": False, "b_last_col_doll": True,
                    "b_first_row_at": False, "b_first_row_doll": True,
                    "b_last_row_at": True, "b_last_row_doll": False,
                },
                "result": "@A$1:$B@2"
                # "exception": "The address 'abc' is not recognized as an address, and its pattern could not be "
                #             "determined."
            },
            {
                "label": "TC 03 - set of columns",
                "parameters": {
                    "str_address": "A:B",
                    "b_first_col_at": False, "b_first_col_doll": True,
                    "b_last_col_at": True, "b_last_col_doll": False,
                    # "b_first_row_at": True, "b_first_row_doll": True,
                    # "b_last_row_at": True, "b_last_row_doll": True,
                },
                "result": "$A:@B"
                # "exception": "The address 'abc' is not recognized as an address, and its pattern could not be "
                #             "determined."
            },
            {
                "label": "TC 04 - set of rows",
                "parameters": {
                    "str_address": "1:3",
                    # "b_first_col_at": False, "b_first_col_doll": True,
                    # "b_last_col_at": True, "b_last_col_doll": False,
                    "b_first_row_at": False, "b_first_row_doll": True,
                    "b_last_row_at": True, "b_last_row_doll": False,
                },
                "result": "$1:@3"
                # "exception": "The address 'abc' is not recognized as an address, and its pattern could not be "
                #             "determined."
            },
            {
                "label": "TC 05 - cell",
                "parameters": {
                    "str_address": "A1",
                    "b_first_col_at": False, "b_first_col_doll": True,
                    # "b_last_col_at": True, "b_last_col_doll": False,
                    "b_first_row_at": True, "b_first_row_doll": False,
                    # "b_last_row_at": True, "b_last_row_doll": False,
                },
                "result": "$A@1"
                # "exception": "The address 'abc' is not recognized as an address, and its pattern could not be "
                #             "determined."
            },
            {
                "label": "TC 06 - no impact",
                "parameters": {
                    "str_address": "A1:B2",
                    # "b_first_col_at": True, "b_first_col_doll": True,
                    # "b_last_col_at": True, "b_last_col_doll": True,
                    # "b_first_row_at": True, "b_first_row_doll": True,
                    # "b_last_row_at": True, "b_last_row_doll": True,
                },
                "result": "A1:B2"
                # "exception": "The address 'abc' is not recognized as an address, and its pattern could not be "
                #             "determined."
            },
            {
                "label": "TC 07 - unset",
                "parameters": {
                    "str_address": "@$A@$1:@$B@$2",
                    "b_first_col_at": False,
                    # "b_first_col_doll": False,
                    # "b_last_col_at": False,
                    "b_last_col_doll": False,
                    "b_first_row_at": False,
                    # "b_first_row_doll": False,
                    # "b_last_row_at": False,
                    # "b_last_row_doll": False,
                },
                "result": "$A$1:@B@$2"
                # "exception": "The address 'abc' is not recognized as an address, and its pattern could not be "
                #             "determined."
            },
            {
                "label": "TC 0A - unexpected address",
                "parameters": {
                    "str_address": " A1",
                    "b_first_col_at": False, "b_first_col_doll": True,
                    # "b_last_col_at": True, "b_last_col_doll": False,
                    "b_first_row_at": True, "b_first_row_doll": False,
                    # "b_last_row_at": True, "b_last_row_doll": False,
                },
                # "result": "$A@1"
                "exception": "The address ' A1' is not recognized as an address, and it cannot be tokenized."
            },
            {
                "label": "TC 0B - unexpected change",
                "parameters": {
                    "str_address": "A1",
                    "b_first_col_at": False, "b_first_col_doll": True,
                    "b_last_col_at": True, "b_last_col_doll": False,
                    "b_first_row_at": True, "b_first_row_doll": False,
                    # "b_last_row_at": True, "b_last_row_doll": False,
                },
                # "result": "$A@1"
                "exception": "Trying to set last_col_at='True' for address 'A1', impossible."
            },
        ]:
            # No exception expected, but a result expected
            if "result" in i_test:

                self.assertEqual(
                    first=i_test["result"],
                    second=CellNotationHelper.transform_address_into_relative_and_absolute_address(
                        **i_test["parameters"]
                    ),
                    msg=f"Error when process test with label '{i_test['label']}'."
                )

            # Exception expected
            else:

                with self.assertRaisesRegex(Exception, i_test["exception"]):

                    CellNotationHelper.transform_address_into_relative_and_absolute_address(
                        **i_test["parameters"]
                    )

        self._logger.debug(f"Function '{stack()[0].filename} - {stack()[0].function}' is returning")

    def test_function__transform_address_into_tokens__transform_tokens_into_address(self):
        self._logger.debug(f"Function '{stack()[0].filename} - {stack()[0].function}' is called")

        for i_test in [
            {
                "label": "TC 01",
                "str_address": "A1:B2",
                "result": {
                    "first_col": "A", "first_col_at": False, "first_col_doll": False,
                    "first_row": "1", "first_row_at": False, "first_row_doll": False,
                    "last_col": "B", "last_col_at": False, "last_col_doll": False,
                    "last_row": "2", "last_row_at": False, "last_row_doll": False,
                },
                # "exception": "The address 'abc' is not recognized as an address, and its pattern could not be "
                #             "determined."
            },
            {
                "label": "TC 02",
                "str_address": "@A$1:$B@$2",
                "result": {
                    "first_col": "A", "first_col_at": True, "first_col_doll": False,
                    "first_row": "1", "first_row_at": False, "first_row_doll": True,
                    "last_col": "B", "last_col_at": False, "last_col_doll": True,
                    "last_row": "2", "last_row_at": True, "last_row_doll": True,
                },
                # "exception": "The address 'abc' is not recognized as an address, and its pattern could not be "
                #             "determined."
            },
            {
                "label": "TC 03 - columns",
                "str_address": "@A:@$B",
                "result": {
                    "first_col": "A", "first_col_at": True, "first_col_doll": False,
                    "first_row": None, "first_row_at": None, "first_row_doll": None,
                    "last_col": "B", "last_col_at": True, "last_col_doll": True,
                    "last_row": None, "last_row_at": None, "last_row_doll": None,
                },
                # "exception": "The address 'abc' is not recognized as an address, and its pattern could not be "
                #             "determined."
            },
            {
                "label": "TC 04 - rows",
                "str_address": "$1:@$2",
                "result": {
                    "first_col": None, "first_col_at": None, "first_col_doll": None,
                    "first_row": "1", "first_row_at": False, "first_row_doll": True,
                    "last_col": None, "last_col_at": None, "last_col_doll": None,
                    "last_row": "2", "last_row_at": True, "last_row_doll": True,
                },
                # "exception": "The address 'abc' is not recognized as an address, and its pattern could not be "
                #             "determined."
            },
            {
                "label": "TC 05 - exception",
                "str_address": " @A$1:$B@$2",
                # "result": {
                #     "first_col": "A", "first_col_at": True, "first_col_doll": False,
                #     "first_row": "1", "first_row_at": False, "first_row_doll": True,
                #     "last_col": "B", "last_col_at": False, "last_col_doll": True,
                #     "last_row": "2", "last_row_at": True, "last_row_doll": True,
                # },
                "exception": "The address ' \\@A\\$1:\\$B\\@\\$2' is not recognized as an address, and it cannot be "
                             "tokenized."
            },
        ]:
            # No exception expected, but a result expected
            if "result" in i_test:

                _dict_tokens: dict = CellNotationHelper.transform_address_into_tokens(
                    str_address=i_test["str_address"]
                )

                # We check the tokenization
                self.assertDictEqual(
                    d1=i_test["result"],
                    d2=_dict_tokens,
                    msg=f"Error when tokenizing test with label '{i_test['label']}'."
                )

                # We check the de-tokenization
                self.assertEqual(
                    first=i_test["str_address"],
                    second=CellNotationHelper.transform_tokens_into_address(
                        dict_tokens=_dict_tokens
                    ),
                    msg=f"Error when process test with label '{i_test['label']}'."
                )

            # Exception expected
            else:

                with self.assertRaisesRegex(Exception, i_test["exception"]):

                    CellNotationHelper.transform_address_into_tokens(
                        str_address=i_test["str_address"]
                    )

        # Eventually, we check exception when de-tokenizing an incorrect dictionary
        with self.assertRaisesRegex(Exception, "Unexpected keys in the parameters dict_tokens.keys\\(\\)='first_col, "
                                               "first_col_at, first_col_doll, first_row, first_row_at, first_row_doll, "
                                               "last_col, last_col_at, last_col_doll, last_row, last_row_at, "
                                               "last_row_doll, jsg'."):
            CellNotationHelper.transform_tokens_into_address(
                dict_tokens={
                    "first_col": None, "first_col_at": None, "first_col_doll": None,
                    "first_row": "1", "first_row_at": False, "first_row_doll": True,
                    "last_col": None, "last_col_at": None, "last_col_doll": None,
                    "last_row": "2", "last_row_at": True, "last_row_doll": True,
                    "jsg": 3
                }
            )

        with self.assertRaisesRegex(Exception, "Unexpected keys in the parameters dict_tokens.keys\\(\\)='first_col, "
                                               "first_col_at, first_col_doll, first_row, first_row_at, first_row_doll, "
                                               "last_col, last_col_at, last_col_doll'."):
            CellNotationHelper.transform_tokens_into_address(
                dict_tokens={
                    "first_col": None, "first_col_at": None, "first_col_doll": None,
                    "first_row": "1", "first_row_at": False, "first_row_doll": True,
                    "last_col": None, "last_col_at": None, "last_col_doll": None,
                    # "last_row": "2", "last_row_at": True, "last_row_doll": True,
                }
            )

        self._logger.debug(f"Function '{stack()[0].filename} - {stack()[0].function}' is returning")



import os
from inspect import stack
from logging import Logger, getLogger
from pathlib import Path
from unittest import TestCase

from com_enovation.toolbox.excel_dashboard.excel_dashboarder import ExcelDashboarder


class TestExcelDashboarder(TestCase):

    _logger: Logger = getLogger(__name__)

    def test_01(self):

        self._logger.debug(f"Function '{stack()[0].filename} - {stack()[0].function}' is called")

        _the_excel_dashboard: ExcelDashboarder = ExcelDashboarder()

        _the_excel_dashboard.excelize(
            p_output_file_path=Path(os.path.join(os.path.dirname(__file__), 'TC01.excel.xlsx')),
            dict_config={
                "sheets": {
                    "test_sheet_name": {
                        "worksheet.freeze_panes": {"row": 1, "col":  1},
                        "worksheet.set_column": [
                            {
                                "first_col": 0,
                                "last_col": 1,
                                "width": 50
                            }
                        ],
                        "widgets": {
                            "test_a_widget_label": {
                                "address": "A1",
                                "widget_id": "https://enovation.com/excel_dashboard/cover_sheet",
                                "config": {
                                    "narrative": "the first text is '{0}' and the second is '{1}'."
                                },
                            }
                        }
                    }
                }
            },
            **{"tokens": ["ici c'est paris", "mais pourquoi donc?"]}
        )

        self._logger.debug(f"Function '{stack()[0].filename} - {stack()[0].function}' is returned")



import os
import unittest

import click
from click.testing import CliRunner, Result
from logging import Logger, getLogger
from inspect import stack

from com_enovation.helper.click.enov_group import EnovGroup


@click.command("set-jsg")
@click.option('--option-test/--no-option-test', type=bool)
@click.argument('arg-test', type=bool)
@click.pass_context
def set_jsg(ctx_context, option_test, arg_test):

    getLogger(__name__).debug(f"Function '{stack()[0].filename} - {stack()[0].function}' is called")

    ctx_context.obj["option_test"] = option_test
    ctx_context.obj["arg_test"] = arg_test

    getLogger(__name__).debug(f"Function '{stack()[0].filename} - {stack()[0].function}' is returning")


@click.command("get-nina")
@click.option('--option-test/--no-option-test', type=bool)
@click.argument('arg-test', type=bool)
@click.pass_context
def get_nina(ctx_context, option_test, arg_test):

    getLogger(__name__).debug(f"Function '{stack()[0].filename} - {stack()[0].function}' is called")

    if "option_test" not in ctx_context.obj:
        raise Exception(f"'option_test' not in ctx_context.obj")

    if ctx_context.obj["option_test"] == option_test:
        raise Exception(f"ctx_context.obj['option_test'] == option_test, while we expect the opposite"
                        f"\n\t- ctx_context.obj['option_test']: '{ctx_context.obj['option_test']}'"
                        f"\n\t- option_test: '{option_test}'")

    if "arg_test" not in ctx_context.obj:
        raise Exception(f"'arg_test' not in ctx_context.obj")

    if ctx_context.obj["arg_test"] == arg_test:
        raise Exception(f"ctx_context.obj['arg_test'] == arg_test, while we expect the opposite"
                        f"\n\t- ctx_context.obj['arg_test']: '{ctx_context.obj['arg_test']}'"
                        f"\n\t- arg_test: '{arg_test}'")

    if "loaded-csv" not in ctx_context.obj:
        raise Exception(f"'loaded-csv' not in ctx_context.obj")

    if len(ctx_context.obj["loaded-csv"].index) != 2:
        raise Exception(f"ctx_context.obj['loaded-csv'] has '{len(ctx_context.obj['loaded-csv'].index)}' rows, while 2 "
                        f"are expected.")

    getLogger(__name__).debug(f"Function '{stack()[0].filename} - {stack()[0].function}' is returning")


nina = EnovGroup(
    str_name="nina",
    lst_commands=[set_jsg, get_nina]
)


class TestExtendedEnov(unittest.TestCase):
    _logger: Logger = getLogger(__name__)

    def test_01_extending_enov_group(self):

        self._logger.debug(f"Function '{stack()[0].filename} - {stack()[0].function}' is called")

        the_runner = CliRunner()

        # First call, all correct
        the_result: Result = the_runner.invoke(
            cli=nina,
            args=[
                '--verbose',

                'set-jsg',
                '--option-test',
                'True',

                'df-load-csv',
                os.path.join(os.path.dirname(__file__), 'test.csv'),
                'loaded-csv',

                'get-nina',
                '--no-option-test',
                'False'
            ],
            catch_exceptions=False,
        )

        self.assertEqual(
            0,
            the_result.exit_code,
            f"Call 1, unexpected exit_code, with stdout: {the_result.stdout}"
        )



import os
from inspect import stack
from unittest import TestCase
from logging import Logger, getLogger

from pandas import DataFrame

from com_enovation.helper.excel_loader import ExcelLoaderBean, ExcelLoader
from com_enovation.helper.pandas_dataframe_typer import PandasDataframeTyper
from com_enovation.toolbox.backlog.bck_compare.bck_comparer import BacklogComparer


class TestFunctionCompare(TestCase):

    _logger: Logger = getLogger(__name__)

    def test_01(self):
        self._logger.debug(f"Function '{stack()[0].filename} - {stack()[0].function}' is called")

        # We load the test data
        _bean_test_data: ExcelLoaderBean = ExcelLoader().load_tables(
            str_path=os.path.join(os.path.dirname(__file__), 'TC01.SimpleSourceTargetAndExpectedResults.xlsx')
        )

        df_the_delta: DataFrame = BacklogComparer().compare(
            df_src=_bean_test_data.dict_tables["src"],
            df_tgt=_bean_test_data.dict_tables["tgt"],
            dict_columns={
                "key columns": {
                    "key 1": {
                        "type": PandasDataframeTyper.str__type__str,
                    },
                    "key 2": {
                        "type": PandasDataframeTyper.str__type__str,
                    },
                },
                "compared columns": {
                    "col 1": {
                        "type": PandasDataframeTyper.str__type__int,
                        #"compare logic": "",
                        #"delta type": ""
                    },
                    "col 2": {
                        "type": PandasDataframeTyper.str__type__str,
                        #"compare logic": "",
                        #"delta type": ""
                    },
                    "col 4": {
                        "type": PandasDataframeTyper.str__type__str,
                        "compare logic": "src.str.len() == tgt.str.len() ",
                        #"delta type": ""
                    },
                }
            }
        )

        df_the_delta: DataFrame = BacklogComparer().compare_into_excel(
            df_src=_bean_test_data.dict_tables["src"],
            df_tgt=_bean_test_data.dict_tables["tgt"],
            dict_columns={
                "key columns": {
                    "key 1": {
                        "type": PandasDataframeTyper.str__type__str,
                    },
                    "key 2": {
                        "type": PandasDataframeTyper.str__type__str,
                    },
                },
                "compared columns": {
                    "col 1": {
                        "type": PandasDataframeTyper.str__type__int,
                        # "compare logic": "",
                        # "delta type": ""
                    },
                    "col 2": {
                        "type": PandasDataframeTyper.str__type__str,
                        # "compare logic": "",
                        # "delta type": ""
                    },
                    "col 4": {
                        "type": PandasDataframeTyper.str__type__str,
                        "compare logic": "src.str.len() == tgt.str.len() ",
                        # "delta type": ""
                    },
                }
            },
            str_path=os.path.join(os.path.dirname(__file__), 'TC01.outputs.xlsx')
        )

        self._logger.debug(f"Function '{stack()[0].filename} - {stack()[0].function}' is returning")



import base64

from dash import dcc, callback, Output, Input, no_update

import dash_bootstrap_components as dbc

from inspect import stack
from logging import Logger, getLogger

from dash import html
from urllib.parse import unquote

from com_enovation.toolbox.dash.session import Session
from com_enovation.toolbox.dash.sheet import Sheet, SheetAsDiv
from com_enovation.toolbox.dash.widget import Widget


class NavigationAsNavbar(Widget, dbc.Navbar):

    _logger: Logger = getLogger(__name__)

    # ##################################################################################################################
    # CONSTANTS
    # ##################################################################################################################

    const_str_default_navbar_brand: str = "eNOVAtion"
    const_str_default_str_logo: str = "images/com-enovation.square.png"

    # The labels in the configuration file
    const_str__config_label__dict_properties: str = "properties"
    const_str__config_label__str_navbar_brand: str = "navbar_brand"
    const_str__config_label__url_logo: str = "logo"
    const_str__config_label__str_title: str = "title"

    # The wid for the default generic widgets
    const_str__wid__div_debug: str = "debug"
    const_str__wid__location_url: str = "url"
    const_str__wid__navbar_header: str = "header"

    """ 
    Return the navigation bar, along with the callback function to load the enov-content__as_main part
    - An image as the logo
    - A brand, which is used as well as the title for the logo
    - Menus to various widgets available.

    The produced HTML is of this kind:
    <nav id="enov-header__as_navbar" class="fixed-top navbar navbar-expand-md navbar-dark bg-dark">
        <div class="nav-item">
            <a href="https://pypi.org/project/com-enovation/">
                <div class="align-items-center row">
                    <div class="col">
                        <img height="30px" src="/assets/images/com-enovation.square.png" title="com.enovation">
                    </div>
                    <div class="col">
                        <span class="navbar-brand">eNOVAtion</span>
                    </div>
                </div>
            </a>
        </div>
        <div class="nav-item">
            <a href="#welcome" class="nav-link">welcome</a>
        </div>
        <div class="nav-item">
            <a href="#explore" class="nav-link">explore</a>
        </div>
        <div class="nav-item">
            <a class="nav-link disabled" disabled="">The default dp_grapher for date predictability module...</a>
        </div>
    </nav>        
    """

    def __init__(
            self,
            dict_sheet: dict[str, Sheet],
            dict_config: dict,
            dict_context: dict,
            str__wid__main_content: str,
            session: Session
    ):

        self._logger.debug(f"Function '{stack()[0].filename} - {stack()[0].function}' is called.")

        _str_the_title: str = dict_config.get(
            self.const_str__config_label__dict_properties, {}
        ).get(self.const_str__config_label__str_title, self.const_str_default_navbar_brand)

        if self.const_str__config_label__url_logo in dict_config.get(
                self.const_str__config_label__dict_properties, {}
        ):
            _img_the_logo: html.Img = html.Img(
                src='data:image/png;base64,{}'.format(
                    base64.b64encode(
                        open(
                            dict_config[
                                self.const_str__config_label__dict_properties
                            ][
                                self.const_str__config_label__url_logo
                            ],
                            'rb'
                        ).read()
                    )
                ),
                height="30px",
                title=_str_the_title
            )
        else:
            _img_the_logo: html.Img = html.Img(
                src=session.dash_application.dash_server.get_asset_url(self.const_str_default_str_logo),
                height="30px",
                title=_str_the_title
            )

        Widget.__init__(
            self,
            dict_config=dict_config,
            dict_context=dict_context,
            session=session
        )

        dbc.Navbar.__init__(
            self,
            children=[
                # To get the url loaded
                dcc.Location(id=self.const_str__wid__location_url),

                dbc.NavItem(
                    html.A(
                        dbc.Row(
                            [
                                dbc.Col(_img_the_logo),
                                dbc.Col(
                                    dbc.NavbarBrand(dict_config.get(
                                        self.const_str__config_label__dict_properties, {}
                                    ).get(
                                        self.const_str__config_label__str_navbar_brand,
                                        self.const_str_default_navbar_brand)
                                    )
                                ),
                            ],
                            align="center",
                        ),
                        href="https://pypi.org/project/com-enovation/"
                    )
                )
            ] + [
                dbc.NavItem(dbc.NavLink(i_widget, href="#"+i_widget))
                for i_widget in dict_sheet
            ] + [
                dbc.NavItem(dbc.NavLink("The default dp_grapher for the date predictability module...", disabled=True))
            ],
            color="dark",
            dark=True,
            id=self.const_str__wid__navbar_header,
            className="fixed-top"
        )

        @callback(
            Output(str__wid__main_content, 'children'),
            Input(self.const_str__wid__location_url, 'hash'),
            prevent_initial_call=True
        )
        def _callback_navigate_and_load_content(url_hash: str) -> html.Div:
            """
            Callback function:
            - When we click in a menu in the navigation bar, it refreshes the page with an updated URL
              "https://xxx/#whatever"
            - When such URL update happens, the widget "url as location" is refreshed
            - Using this refresh, we load the area "content as main" with the new target widget.

            :param url_hash: the link to an element with a specified id within the page
            :return: the widget to load in the area "content as main"
            """
            getLogger(__name__).debug(f"Function '{stack()[0].filename} - {stack()[0].function}' is called.")

            if url_hash == '':
                return no_update

            _dict_the_sheets: dict[str, SheetAsDiv] = self.session.dash_application.dict_sheets

            if unquote(url_hash)[1:] not in _dict_the_sheets:
                raise Exception(f"Unexpected url_hash '{unquote(url_hash)}'.")

            div_the_return: html.Div = _dict_the_sheets[unquote(url_hash)[1:]]

            self.session.debug(
                debug_trace=f"Click navigation menus to update main content: {unquote(url_hash)}, returned an "
                            f"object of type '{type(div_the_return)}'"
            )

            getLogger(__name__).debug(f"Function '{stack()[0].filename} - {stack()[0].function}' is returning.")

            return div_the_return

        self._logger.debug(f"Function '{stack()[0].filename} - {stack()[0].function}' is returning.")



import json
import os
from inspect import stack
from logging import Logger, getLogger
from unittest import TestCase

from pandas import DataFrame, read_excel

from com_enovation.helper.json_encoder import JSONEncoder


class Test01DataframeJsonifyer(TestCase):

    _logger: Logger = getLogger(__name__)

    def test_01_dataframe_jsonifyer(self):
        self._logger.debug(f"Function '{stack()[0].filename} - {stack()[0].function}' is called.")

        df_the_dataframe: DataFrame = read_excel(
            io=os.path.join(os.path.dirname(__file__), 'TC_A_00.DataframeToJsonify.xls'),
            dtype="object",
        )

        str_the_json: str = json.dumps(
            obj=df_the_dataframe,
            cls=JSONEncoder
        )

        self.assertEqual(
            first='{"col1_str": {"0": "aaa", "1": "bbb", "2": "ccc", "3": "ddd", "4": "eee", "5": NaN}, "col2_obj": {"0'
                  '": "aaa", "1": "2021-03-12T00:00:00", "2": 13.2, "3": 13, "4": NaN, "5": "qdsf"}, "col3_date": {"0":'
                  ' 44217, "1": 44248, "2": 44276, "3": 44307, "4": 44337, "5": NaN}, "col4_float": {"0": 1.1, "1": 1.2'
                  ', "2": 1.3333333, "3": 1, "4": 3.51, "5": NaN}, "col5_int": {"0": 1, "1": 2, "2": 3, "3": 3, "4": 4,'
                  ' "5": NaN}, "Col 6 int": {"0": 1, "1": 2, "2": "3,5", "3": 4.4, "4": "dd", "5": NaN}}',
            second=str_the_json,
            msg="Unexpected json..."
        )

        self._logger.debug(f"Function '{stack()[0].filename} - {stack()[0].function}' is returning.")



from inspect import stack
from logging import Logger, getLogger
from unittest import TestCase

from com_enovation.toolbox.excel_dashboard.cell_notation_helper import CellNotationHelper


class TestCellNotationHelper_TranslateFormulaToReference(TestCase):

    _logger: Logger = getLogger(__name__)

    def test__function__translate_formula_to_reference(self):
        self._logger.debug(f"Function '{stack()[0].filename} - {stack()[0].function}' is called")

        _dict_test_data: dict = {
            '=function("")': '=function("")',
            '=@C3': '=D3',
            '=@$C3': '=$D3',
            '=function("@B3")': '=function("@B3")',
            '=function(@C3)': '=function(D3)',
            '=function([@C3])': '=function([@C3])',
            '=function(C:@C)': '=function(C:D)',
            '=function(C@3:@C10)': '=function(C5:D10)',
            '=RIGHT(@C@3,1)="A"': '=RIGHT(D5,1)="A"'
        }

        for k, v in _dict_test_data.items():
            _str_result: str = CellNotationHelper.translate_formula_to_reference(
                str_reference_address="B3",
                str_formula=k
            )
            self.assertEqual(
                first=v,
                second=_str_result,
                msg=f"Function CellNotationHelper.get_absolute_from_relative_address(str_reference_address='B3', "
                    f"str_relative_address='{k}') returned '{_str_result}' while we expected '{v}'."
            )

        self._logger.debug(f"Function '{stack()[0].filename} - {stack()[0].function}' is returning")

    def test__function__translate_formula_to_reference__tokenize(self):

        _dict_test_data: dict = {
            '=function("")': ["=function(", r'""', ")"],
            '=""&function("")': ["=", r'""', "&function(", r'""', ")"],
            '="with an escaped "" double quote" &function("")': ["=", r'"with an escaped "" double quote"',
                                                                 " &function(", r'""', ")"],
        }

        for k_expression, v_expected_tokens in _dict_test_data.items():

            _lst_the_tokens: list[str] = CellNotationHelper._translate_formula_to_reference__tokenize(
                str_formula=k_expression
            )

            self.assertListEqual(
                list1=v_expected_tokens,
                list2=_lst_the_tokens,
                msg=f"Error when processing the expression '{k_expression}'."
            )

    def test__function__translate_formula_to_reference__pattern_a1(self):
        _dict_test_data: dict = {
            "C5": "C5",
            "@C@5": "D7",
            "=function(C5)": "=function(C5)",

            "[@C5]": "[@C5]",
        }

        for k_expression, v_expected_transformed in _dict_test_data.items():

            _str_the_transformed: str = CellNotationHelper._translate_formula_to_reference__pattern_a1(
                str_formula_token=k_expression,
                str_reference_address="B3"
            )

            self.assertEqual(
                first=v_expected_transformed,
                second=_str_the_transformed,
                msg=f"Error when processing the expression '{k_expression}'."
            )

    def test__function__translate_formula_to_reference__pattern_aa(self):

        _dict_test_data: dict = {
            "C5": "C5",
            "@C@5": "@C@5",
            "=function(C5)": "=function(C5)",
            "@C5": "@C5",

            "C:C": "C:C",
            "@C:D": "D:D",
            "=function(C:@C)": "=function(C:D)",

            "[@C[label]]": "[@C[label]]",
            "=function([@C[label])": "=function([@C[label])",
        }

        for k_expression, v_expected_transformed in _dict_test_data.items():

            _str_the_transformed: str = CellNotationHelper._translate_formula_to_reference__pattern_aa(
                str_formula_token=k_expression,
                str_reference_address="B3"
            )

            self.assertEqual(
                first=v_expected_transformed,
                second=_str_the_transformed,
                msg=f"Error when processing the expression '{k_expression}'."
            )

    def test__function__translate_formula_to_reference__pattern_11(self):

        _dict_test_data: dict = {
            "C5": "C5",
            "@C@5": "@C@5",
            "=function(C5)": "=function(C5)",
            "@C5": "@C5",

            "C:C": "C:C",
            "@C:C": "@C:C",
            "=function(C:@C)": "=function(C:@C)",

            "5:5": "5:5",
            "@5:7": "7:7",
            "=function(5:@5)": "=function(5:7)",
        }

        for k_expression, v_expected_transformed in _dict_test_data.items():

            _str_the_transformed: str = CellNotationHelper._translate_formula_to_reference__pattern_11(
                str_formula_token=k_expression,
                str_reference_address="B3"
            )

            self.assertEqual(
                first=v_expected_transformed,
                second=_str_the_transformed,
                msg=f"Error when processing the expression '{k_expression}'."
            )



import os
import unittest
from pathlib import Path
from logging import Logger, getLogger
from inspect import stack
from pandas import DataFrame, read_excel
from pandas.testing import assert_frame_equal

from com_enovation.toolbox.data_handler.data_modifier.data_modification_feature_factory \
    import DataModificationFeatureFactory


class TestDataModificationFeatureFactory_ReplaceMissingValues(unittest.TestCase):
    _logger: Logger = getLogger(__name__)

    def test_replace_missing_values(self):
        self._logger.debug(f"Function '{stack()[0].filename} - {stack()[0].function}' is called")

        # We load the data extract
        df_data_extract: DataFrame = read_excel(Path(os.path.join(os.path.dirname(__file__),
                                                                  '01.data_extract.xlsx')))

        # Run 1: no missing value in the column, therefore no modification
        df_the_modified_dataframe: DataFrame = DataModificationFeatureFactory(). \
            get_data_modification_feature__replace_missing_values(
            dict_columns={
                "Jsg - ID": "ABC",
            },
            b_copy=True
        )(df_data_extract)

        assert_frame_equal(
            left=df_data_extract,
            right=df_the_modified_dataframe,
            check_exact=True,
            obj="Run 1 - We did not modify any missing value..."
        )

        # Run 2: missing values are defaulted
        df_the_modified_dataframe: DataFrame = DataModificationFeatureFactory(). \
            get_data_modification_feature__replace_missing_values(
            dict_columns={
                "Jsg - ID": "ABC",
                "Jsg - Null Values A": "defaulted missing value",
                "Jsg - Null Values B": 7.321
            },
            b_copy=True
        )(df_data_extract)

        self.assertEqual(
            first=len(
                df_the_modified_dataframe[
                    df_the_modified_dataframe["Jsg - Null Values A"] == "defaulted missing value"
                ].index
            ),
            second=5,
            msg="Run 2, check 1 - We expect 5 lines to be modified in column A..."
        )

        self.assertEqual(
            first=len(
                df_the_modified_dataframe[
                    df_the_modified_dataframe["Jsg - Null Values B"] == 7.321
                ].index
            ),
            second=1,
            msg="Run 2, check 2 - We expect 3 lines to be modified in column B..."
        )

        self._logger.debug(f"Function '{stack()[0].filename} - {stack()[0].function}' is returning")



from datetime import datetime
from inspect import stack
from unittest import TestCase
from logging import Logger, getLogger

from pandas import DataFrame
from pandas._testing import assert_frame_equal

from com_enovation.toolbox import PredictabilityBean
from com_enovation.toolbox.predictability.dp_date_predictability.dp_computer import DatePredictabilityComputer


class TestFunctionComputeHistoricalDatePredictability(TestCase):
    """
    Test public fonctions that compute historical date predictability, with or without resampling
    """

    _logger: Logger = getLogger(__name__)

    df_the_test_date: DataFrame = DataFrame(
        columns=[
            DatePredictabilityComputer.str__input__column_label__key,
            DatePredictabilityComputer.str__input__column_label__date,
            DatePredictabilityComputer.str__input__column_label__measure,
            DatePredictabilityComputer.str__output__column_label__predictability,
        ],
        data=[
            ['abc', '2021-4-26', '2021-5-22', 1],
            ['abc', '2021-4-27', '2021-6-5', 0.65],
            ['abc', '2021-4-30', '2021-6-3', 0.879364514890831],
            ['abc', '2021-5-1', '2021-6-4', 0.91139492876335],
            ['abc', '2021-5-6', '2021-5-29', 0.778029804255193],
            ['abc', '2021-5-7', '2021-5-25', 0.620045338991703],
            ['abc', '2021-5-9', '2021-5-21', 0.41092854342655],
            ['abc', '2021-5-11', '2021-5-19', 0.305391170303002],
            ['abc', '2021-5-13', '2021-5-16', 0.106022408963585],
            ['abc', '2021-5-16', '2021-5-17', 0.261746031746032],
            ['abc', '2021-5-18', '2021-5-21', 0.452146808850176],
            ['abc', '2021-5-23', '2021-5-14', 0.0252057613168724],
            ['abc', '2021-5-25', '2021-5-27', 0.45598774482143],
            ['abc', '2021-5-29', '2021-5-20', 0.304694618503066],
            ['abc', '2021-6-6', '2021-5-19', 0.3877575869183],
        ]
    )

    df_the_expected_stats: DataFrame = DataFrame(
        columns=[
            DatePredictabilityComputer.str__input__column_label__key,
            DatePredictabilityComputer.str__statistics__column_label__measure_count,
            DatePredictabilityComputer.str__statistics__column_label__date_first,
            DatePredictabilityComputer.str__statistics__column_label__date_last,
            DatePredictabilityComputer.str__statistics__column_label__measure_min,
            DatePredictabilityComputer.str__statistics__column_label__measure_max,
            DatePredictabilityComputer.str__statistics__column_label__measure_first,
            DatePredictabilityComputer.str__statistics__column_label__measure_last,
            DatePredictabilityComputer.str__statistics__column_label__predictability_last
        ],
        data=[[
            'abc',                  # key
            15,                     # measure_count
            datetime(2021, 4, 26),  # date_first
            datetime(2021, 6, 6),   # date_last
            datetime(2021, 5, 14),  # measure_min
            datetime(2021, 6, 5),   # measure_max
            datetime(2021, 5, 22),  # measure_first
            datetime(2021, 5, 19),  # measure_last
            0.3877575869183         # predictability_last
        ]]
    )

    def test_01_resampling(self):
        self._logger.debug(f"Function '{stack()[0].filename} - {stack()[0].function}' is called")

        # 1. We call the function
        the_bean: PredictabilityBean = DatePredictabilityComputer().compute_historical_date_predictability(
            df_measures=self.df_the_test_date[[
                DatePredictabilityComputer.str__input__column_label__key,
                DatePredictabilityComputer.str__input__column_label__date,
                DatePredictabilityComputer.str__input__column_label__measure,
            ]],
            b_resample=True
        )

        # 2. We check the historical data frame
        df_the_historical: DataFrame = the_bean.df_historical

        for t_i_row in self.df_the_test_date.itertuples():
            str_the_key: str = getattr(t_i_row, DatePredictabilityComputer.str__input__column_label__key)
            dt_the_date: datetime = getattr(t_i_row, DatePredictabilityComputer.str__input__column_label__date)

            f_the_expected_predictability: float = getattr(
                t_i_row,
                DatePredictabilityComputer.str__output__column_label__predictability
            )

            f_the_computed_predictability: float = df_the_historical[
                (df_the_historical[DatePredictabilityComputer.str__input__column_label__key] == str_the_key)
                & (df_the_historical[DatePredictabilityComputer.str__input__column_label__date] == dt_the_date)
            ][DatePredictabilityComputer.str__output__column_label__predictability].iloc[0]

            self.assertAlmostEqual(
                first=f_the_expected_predictability,
                second=f_the_computed_predictability,
                msg=f"Incorrect computation '{f_the_computed_predictability}' vs expected "
                    f"'{f_the_expected_predictability}' for record:"
                    f"\n\t- key: '{str_the_key}'"
                    f"\n\t- date: '{dt_the_date}'"
            )

        # 3. We check the statistics
        assert_frame_equal(
            left=self.df_the_expected_stats,
            right=the_bean.df_by_key,
            obj=f"The computation did not return expected statistics",
            atol=0.00001,
            check_like=True  # To exclude index, and not to care about column ordering
        )

        self._logger.debug(f"Function '{stack()[0].filename} - {stack()[0].function}' is returning")

    def test_02_no_resampling(self):
        self._logger.debug(f"Function '{stack()[0].filename} - {stack()[0].function}' is called")

        # 1. We call the function
        the_bean: PredictabilityBean = DatePredictabilityComputer().compute_historical_date_predictability(
            df_measures=self.df_the_test_date[[
                DatePredictabilityComputer.str__input__column_label__key,
                DatePredictabilityComputer.str__input__column_label__date,
                DatePredictabilityComputer.str__input__column_label__measure,
            ]],
            b_resample=False
        )

        # 2. We check the historical data frame
        df_the_historical: DataFrame = the_bean.df_historical

        for t_i_row in self.df_the_test_date.itertuples():
            str_the_key: str = getattr(t_i_row, DatePredictabilityComputer.str__input__column_label__key)
            dt_the_date: datetime = getattr(t_i_row, DatePredictabilityComputer.str__input__column_label__date)

            f_the_expected_predictability: float = getattr(
                t_i_row,
                DatePredictabilityComputer.str__output__column_label__predictability
            )

            f_the_computed_predictability: float = df_the_historical[
                (df_the_historical[DatePredictabilityComputer.str__input__column_label__key] == str_the_key)
                & (df_the_historical[DatePredictabilityComputer.str__input__column_label__date] == dt_the_date)
            ][DatePredictabilityComputer.str__output__column_label__predictability].iloc[0]

            self.assertAlmostEqual(
                first=f_the_expected_predictability,
                second=f_the_computed_predictability,
                msg=f"Incorrect computation '{f_the_computed_predictability}' vs expected "
                    f"'{f_the_expected_predictability}' for record:"
                    f"\n\t- key: '{str_the_key}'"
                    f"\n\t- date: '{dt_the_date}'"
            )

        # 3. We check the statistics
        assert_frame_equal(
            left=self.df_the_expected_stats,
            right=the_bean.df_by_key,
            obj=f"The computation did not return expected statistics",
            atol=0.00001,
            check_like=True  # To exclude index, and not to care about column ordering
        )

        self._logger.debug(f"Function '{stack()[0].filename} - {stack()[0].function}' is returning")



import logging
from pathlib import Path

import click

from com_enovation.helper.excel.xls_refresh_data_connections import XlsRefreshDataConnections

_logger: logging.Logger = logging.getLogger(__name__)


@click.command('xls-refresh')
@click.argument('file', type=click.Path(exists=True))
@click.option('--save_as', default=None, type=click.Path(), help='To persist the refresh excel file as.')
def xls_refresh_data_connections(
        file: Path,
        save_as: Path
):
    """
    Command to open an Excel FILE, refresh all queries and connections, before persisting the file or SAVE-AS another
    path.

    Note: when crashing, a zombie process might survive:
    - ps | grep EXCEL
    - kill xxx
    """

    XlsRefreshDataConnections().refresh_data_connections(
        p_file=Path(file),
        p_save_as=Path(save_as)
    )



import os
import shutil
import unittest
from inspect import stack
from logging import getLogger, Logger
from sys import platform
from unittest import TestCase

import click
from click.testing import CliRunner

from com_enovation.enov import enov

from com_enovation.helper.excel_loader import ExcelLoader, ExcelLoaderBean
from pandas import DataFrame

from pandas.testing import assert_frame_equal


@unittest.skipUnless(platform.startswith("win"), f"CLASS - Do run on Windows only, and not on {platform}")
class TestClick_XlsConnectionsToDf(TestCase):

    _logger: Logger = getLogger(__name__)

    @unittest.skipUnless(platform.startswith("win"), f"FN - Do run on Windows only, and not on {platform}")
    def test_01_no_persist(self):
        self._logger.debug(f"Function '{stack()[0].filename} - {stack()[0].function}' is called.")

        try:
            os.remove(os.path.join(os.path.dirname(__file__), "./02.ConnectedData.persisted.xlsx"))
        except FileNotFoundError:
            pass

        runner = CliRunner()

        enov.add_command(check_data)

        # noinspection PyTypeChecker
        result = runner.invoke(
            cli=enov,
            args=[
                # To get all the logs...
                '--verbose',

                'xls-connection-to-df',
                '--no-persist',
                os.path.join(os.path.dirname(__file__), "./02.ConnectedData.xlsx"),
                "JSG",

                'check_data',
                'TEST 01. no persist'
            ]
        )

        self.assertEqual(
            first=0,
            second=result.exit_code,
            msg="TEST 01: " + str(result.exception)
        )

    @unittest.skipUnless(platform.startswith("win"), f"FN - Do run on Windows only, and not on {platform}")
    def test_02_persist(self):
        self._logger.debug(f"Function '{stack()[0].filename} - {stack()[0].function}' is called.")

        try:
            os.remove(os.path.join(os.path.dirname(__file__), "./02.ConnectedData.persisted.xlsx"))
        except FileNotFoundError:
            pass

        runner = CliRunner()

        enov.add_command(check_data)

        # We need to copy the file that will be refreshed, as it will be persisted!
        shutil.copyfile(
            src=os.path.join(os.path.dirname(__file__), "./02.ConnectedData.xlsx"),
            dst=os.path.join(os.path.dirname(__file__), "./02.ConnectedData.persisted.xlsx")
        )

        # noinspection PyTypeChecker
        result = runner.invoke(
            cli=enov,
            args=[
                # To get all the logs...
                '--verbose',

                'xls-connection-to-df',
                # '--persist',
                os.path.join(os.path.dirname(__file__), "./02.ConnectedData.persisted.xlsx"),
                "JSG",

                'check_data',
                'TEST 02. persist'
            ]
        )

        self.assertEqual(
            first=0,
            second=result.exit_code,
            msg="TEST 02: " + str(result.exception)
        )


@unittest.skipUnless(platform.startswith("win"), f"FN - Do run on Windows only, and not on {platform}")
@click.command('check_data')
@click.argument('step', type=str)
@click.pass_context
def check_data(ctx_context, step):
    """
    Function that checks data were effectively loaded...
    """

    # We get source data was to later reconcile
    _excel_source_tables: ExcelLoaderBean = ExcelLoader().load_tables(
        str_path=os.path.join(os.path.dirname(__file__), "./01.SourceData.xlsx")
    )

    # We check data was not persisted in 02.ConnectedData
    _excel_connected_tables: ExcelLoaderBean = ExcelLoader().load_tables(
        str_path=os.path.join(os.path.dirname(__file__), "./02.ConnectedData.xlsx")
    )
    for i_table in ["JSG_TABLE_1", "BCH_TABLE_1", "BCH_TABLE_2"]:
        if len(_excel_connected_tables.dict_tables[i_table+"query"].index) != 1:
            raise Exception(f"Table {i_table} has {len(_excel_connected_tables.dict_tables[i_table+'query'].index)}, "
                            f"while only one is expected.")

    # We check data loaded in context
    # Note: we only import the dependency here, as this import can only work on Windows!
    from com_enovation.helper.excel.xls_data_connection_to_df import ExcelTables
    _obj_excel_tables: ExcelTables = ctx_context.obj["JSG"]

    for i_table in ["JSG_TABLE_1", "BCH_TABLE_1", "BCH_TABLE_2"]:

        _src: DataFrame = _excel_source_tables.dict_tables[i_table]
        _refreshed: DataFrame = _obj_excel_tables.dict_tables[i_table+'query']

        assert_frame_equal(
            left=_src,
            right=_refreshed,
            obj=f"{i_table} not reconciling",
            check_dtype=False
        )

    if step == 'TEST 02. persist':

        # We check data from 02.ConnectedData
        _excel_tables: ExcelLoaderBean = ExcelLoader().load_tables(
            str_path=os.path.join(os.path.dirname(__file__), "./02.ConnectedData.persisted.xlsx")
        )
        for i_table in ["JSG_TABLE_1", "BCH_TABLE_1", "BCH_TABLE_2"]:

            _src: DataFrame = _excel_source_tables.dict_tables[i_table]
            _refreshed: DataFrame = _excel_tables.dict_tables[i_table+'query']

            assert_frame_equal(
                left=_src,
                right=_refreshed,
                obj=f"{i_table} not reconciling",
                check_dtype=False
            )



from inspect import stack
from logging import Logger, getLogger
from unittest import TestCase

import numpy
import pandas
from pandas import DataFrame

from com_enovation.helper.pandas_dataframe_sampler import PandasDataframeSampler
from pandas.testing import assert_frame_equal

from com_enovation.helper.pandas_dataframe_typer import PandasDataframeTyper


class Test01DataframeSampler(TestCase):

    _logger: Logger = getLogger(__name__)

    def test_01_single_record_per_key(self):
        self._logger.debug(f"Function '{stack()[0].filename} - {stack()[0].function}' is called.")

        df_the_data: DataFrame = DataFrame(
            columns=[
                "col 1", "col 2", "col 3", "col 4"
            ],
            data=[
                ["aa_a", "bbb_b", "cccc_c", "dddd_d"],
                ["aa_a", "bbb_c", "cccc_d", "dddd_d"],
                ["aa_a", "bbb_d", "cccc_d", "dddd_e"],
                ["aa_a", "bbb_e", "cccc_d", "dddd_f"],

            ]
        )

        # Combination 1: 2 key columns, 2 value columns
        assert_frame_equal(
            left=df_the_data,
            right=PandasDataframeSampler.compress(
                df_measures=df_the_data,
                lst_key_columns=["col 1", "col 2"],
                lst_value_columns=["col 3", "col 4"],
                b_keep_last=True,
            ),
            obj=f"Call 1",
            atol=0.00001,
            check_like=True  # To exclude index, and not to care about column ordering
        )

        # Combination 2: only one key column, 3 value columns
        assert_frame_equal(
            left=df_the_data,
            right=PandasDataframeSampler.compress(
                df_measures=df_the_data,
                lst_key_columns=["col 1"],
                lst_value_columns=["col 2", "col 3", "col 4"],
                b_keep_last=True,
            ),
            obj=f"Call 2",
            atol=0.00001,
            check_like=True  # To exclude index, and not to care about column ordering
        )

        # Combination 3: no key column (empty list), only value columns
        assert_frame_equal(
            left=df_the_data,
            right=PandasDataframeSampler.compress(
                df_measures=df_the_data,
                lst_key_columns=[],
                lst_value_columns=["col 1", "col 2", "col 3", "col 4"],
                b_keep_last=True,
            ),
            obj=f"Call 3",
            atol=0.00001,
            check_like=True  # To exclude index, and not to care about column ordering
        )

        # Combination 4: no key column (none value), only value columns
        assert_frame_equal(
            left=df_the_data,
            right=PandasDataframeSampler.compress(
                df_measures=df_the_data,
                lst_value_columns=["col 1", "col 2", "col 3", "col 4"],
                b_keep_last=True,
            ),
            obj=f"Call 4",
            atol=0.00001,
            check_like=True  # To exclude index, and not to care about column ordering
        )

        self._logger.debug(f"Function '{stack()[0].filename} - {stack()[0].function}' is returning.")

    def test_02_drop_last(self):
        self._logger.debug(f"Function '{stack()[0].filename} - {stack()[0].function}' is called.")

        df_the_data: DataFrame = DataFrame(
            columns=[
                "col 1", "col 2", "col 3", "col 4", "expected result"
            ],
            data=[
                ["aa_a", "bbb_b", "cccc_c", "dddd_d", "filter in"],

                ["aa_a", "bbb_c", "cccc_d", "dddd_d", "filter in"],

                ["aa_a", "bbb_d", "cccc_d", "dddd_e", "filter in"],
                ["aa_a", "bbb_d", "cccc_d", "dddd_e", "filter out"],
                ["aa_a", "bbb_d", "cccc_d", "dddd_e", "filter out"],
                ["aa_a", "bbb_d", "cccc_d", "dddd_e", "filter out"],

                ["aa_a", "bbb_e", "cccc_d", "dddd_f", "filter in"],
                ["aa_a", "bbb_e", "cccc_d", "dddd_f", "filter out"],
                ["aa_a", "bbb_e", "cccc_d", "dddd_f", "filter out"],
            ],
        )

        df_the_return: DataFrame = PandasDataframeSampler.compress(
            df_measures=df_the_data[[i_col for i_col in df_the_data.columns if i_col.startswith("col")]],
            lst_key_columns=["col 1", "col 2"],
            lst_value_columns=["col 3", "col 4"],
            b_keep_last=False,
        )

        assert_frame_equal(
            left=df_the_data[
                df_the_data["expected result"] == "filter in"
            ][df_the_return.columns].reset_index(drop=True),
            right=df_the_return.reset_index(drop=True),
            obj=f"Call 1",
            atol=0.00001,
            check_like=True,  # To exclude index, and not to care about column ordering
            check_names=False,
        )

        self._logger.debug(f"Function '{stack()[0].filename} - {stack()[0].function}' is returning.")

    def test_03_keep_last(self):
        self._logger.debug(f"Function '{stack()[0].filename} - {stack()[0].function}' is called.")

        df_the_data: DataFrame = DataFrame(
            columns=[
                "col 1", "col 2", "col 3", "col 4", "expected result"
            ],
            data=[
                ["aa_a", "bbb_b", "cccc_c", "dddd_d", "filter in"],

                ["aa_a", "bbb_c", "cccc_d", "dddd_d", "filter in"],

                ["aa_a", "bbb_d", "cccc_d", "dddd_e", "filter in"],
                ["aa_a", "bbb_d", "cccc_d", "dddd_e", "filter out"],
                ["aa_a", "bbb_d", "cccc_d", "dddd_e", "filter out"],
                ["aa_a", "bbb_d", "cccc_d", "dddd_e", "filter in"],

                ["aa_a", "bbb_e", "cccc_d", "dddd_f", "filter in"],
                ["aa_a", "bbb_e", "cccc_d", "dddd_f", "filter out"],
                ["aa_a", "bbb_e", "cccc_d", "dddd_f", "filter in"],

                ["aa_a", "bbb_f", "cccc_d", "dddd_g", "filter in"],
                ["aa_a", "bbb_f", "cccc_d", "dddd_g", "filter in"],
            ],
        )

        df_the_return: DataFrame = PandasDataframeSampler.compress(
            df_measures=df_the_data[[i_col for i_col in df_the_data.columns if i_col.startswith("col")]],
            lst_key_columns=["col 1", "col 2"],
            lst_value_columns=["col 3", "col 4"],
            b_keep_last=True,
        )

        assert_frame_equal(
            left=df_the_data[
                df_the_data["expected result"] == "filter in"
            ][df_the_return.columns].reset_index(drop=True),
            right=df_the_return.reset_index(drop=True),
            obj=f"Call 1",
            atol=0.00001,
            check_like=True,  # To exclude index, and not to care about column ordering
            check_names=False,
        )

        self._logger.debug(f"Function '{stack()[0].filename} - {stack()[0].function}' is returning.")

    def test_04_full_example_no_keep_last(self):
        self._logger.debug(f"Function '{stack()[0].filename} - {stack()[0].function}' is called.")

        df_the_data: DataFrame = DataFrame(
            columns=[
                "key 1", "key 2", "order 1", "order 2", "val 1", "val 2", "expected result"
            ],
            data=[
                # Universe 1
                ["k1a", "k2a", "o1a", "o2a", "v1a", "v2a", "filter in"],
                ["k1a", "k2a", "o1a", "o2b", "v1a", "v2a", "filter out"],
                ["k1a", "k2a", "o1b", "o2a", "v1b", "v2a", "filter in"],
                ["k1a", "k2a", "o1b", "o2b", "v1a", "v2a", "filter in"],
                ["k1a", "k2a", "o1b", "o2c", "v1a", "v2a", "filter out"],
                ["k1a", "k2a", "o1c", "o2a", "v1a", "v2a", "filter out"],
                ["k1a", "k2a", "o1c", "o2b", "v1a", "v2a", "filter out"],
            ]
        )

        df_the_return: DataFrame = PandasDataframeSampler.compress(
            df_measures=df_the_data[[i_col for i_col in df_the_data.columns if i_col != "expected result"]],
            lst_key_columns=["key 1", "key 2"],
            lst_ordering_columns=["order 1", "order 2"],
            lst_value_columns=["val 1", "val 2"],
            b_keep_last=False,
        )

        assert_frame_equal(
            left=df_the_data[
                df_the_data["expected result"] == "filter in"
            ][df_the_return.columns].reset_index(drop=True),
            right=df_the_return.reset_index(drop=True),
            obj=f"Call 1",
            atol=0.00001,
            check_like=True  # To exclude index, and not to care about column ordering
        )

        self._logger.debug(f"Function '{stack()[0].filename} - {stack()[0].function}' is returning.")

    def test_05_full_example_keep_last(self):
        self._logger.debug(f"Function '{stack()[0].filename} - {stack()[0].function}' is called.")

        df_the_data: DataFrame = DataFrame(
            columns=[
                "key 1", "key 2", "order 1", "order 2", "val 1", "val 2", "expected result"
            ],
            data=[
                # Universe 1
                ["k1a", "k2a", "o1a", "o2a", "v1a", "v2a", "filter in"],
                ["k1a", "k2a", "o1a", "o2b", "v1a", "v2a", "filter out"],
                ["k1a", "k2a", "o1b", "o2a", "v1b", "v2a", "filter in"],
                ["k1a", "k2a", "o1b", "o2b", "v1a", "v2a", "filter in"],
                ["k1a", "k2a", "o1b", "o2c", "v1a", "v2a", "filter out"],
                ["k1a", "k2a", "o1c", "o2a", "v1a", "v2a", "filter out"],
                ["k1a", "k2a", "o1c", "o2b", "v1a", "v2a", "filter in"],
            ]
        )

        df_the_return: DataFrame = PandasDataframeSampler.compress(
            df_measures=df_the_data[[i_col for i_col in df_the_data.columns if i_col != "expected result"]],
            lst_key_columns=["key 1", "key 2"],
            lst_ordering_columns=["order 1", "order 2"],
            lst_value_columns=["val 1", "val 2"],
            b_keep_last=True,
        )

        assert_frame_equal(
            left=df_the_data[
                df_the_data["expected result"] == "filter in"
            ][df_the_return.columns].reset_index(drop=True),
            right=df_the_return.reset_index(drop=True),
            obj=f"Call 1",
            atol=0.00001,
            check_like=True  # To exclude index, and not to care about column ordering
        )

        self._logger.debug(f"Function '{stack()[0].filename} - {stack()[0].function}' is returning.")

    def test_06_full_example_reordered_keep_last(self):
        self._logger.debug(f"Function '{stack()[0].filename} - {stack()[0].function}' is called.")

        # Same data set as previously...
        df_the_data: DataFrame = DataFrame(
            columns=[
                "key 1", "key 2", "order 1", "order 2", "val 1", "val 2", "expected result"
            ],
            data=[
                # Universe 1
                ["k1a", "k2a", "o1b", "o2a", "v1b", "v2a", 2],   # 3 - in
                ["k1a", "k2a", "o1b", "o2c", "v1a", "v2a", -1],  # 5 - out
                ["k1a", "k2a", "o1a", "o2a", "v1a", "v2a", 1],   # 1 - in
                ["k1a", "k2a", "o1c", "o2b", "v1a", "v2a", 4],   # 6 - in
                ["k1a", "k2a", "o1c", "o2a", "v1a", "v2a", -1],  # 5 - out
                ["k1a", "k2a", "o1a", "o2b", "v1a", "v2a", -1],  # 2 - out
                ["k1a", "k2a", "o1b", "o2b", "v1a", "v2a", 3],   # 4 - in
            ]
        )

        df_the_return: DataFrame = PandasDataframeSampler.compress(
            df_measures=df_the_data[[i_col for i_col in df_the_data.columns if i_col != "expected result"]],
            lst_key_columns=["key 1", "key 2"],
            lst_ordering_columns=["order 1", "order 2"],
            lst_value_columns=["val 1", "val 2"],
            b_keep_last=True,
        )

        assert_frame_equal(
            left=df_the_data[
                df_the_data["expected result"] > 0
            ].sort_values(by="expected result")[df_the_return.columns].reset_index(drop=True),
            right=df_the_return.reset_index(drop=True),
            obj=f"Call 1",
            atol=0.00001,
            check_like=True  # To exclude index, and not to care about column ordering
        )

        self._logger.debug(f"Function '{stack()[0].filename} - {stack()[0].function}' is returning.")

    def test_07_full_example_keep_last_with_null(self):
        self._logger.debug(f"Function '{stack()[0].filename} - {stack()[0].function}' is called.")

        df_the_data: DataFrame = DataFrame(
            columns=[
                "key 1", "key 2", "order 1", "order 2", "val 1", "val 2", "expected result"
            ],
            data=[
                # Universe 1
                ["k1a", "k2a", "o1a", "o2a", "v1a", '2021-1-15', "filter in"],
                ["k1a", "k2a", "o1a", "o2b", "v1a", '2021-1-15', "filter out"],
                ["k1a", "k2a", "o1b", "o2a", "v1b", '2021-1-15', "filter in"],
                ["k1a", "k2a", "o1b", "o2b", "v1a", pandas.NA, "filter in"],
                ["k1a", "k2a", "o1b", "o2c", "v1a", pandas.NaT, "filter out"],
                ["k1a", "k2a", "o1c", "o2a", "v1a", numpy.NAN, "filter out"],
                ["k1a", "k2a", "o1c", "o2b", "v1a", None, "filter in"],

                ["k2a", "k2a", "o1c", "o2b", "v1a", None, "filter in"],

            ]
        )
        df_the_data = PandasDataframeTyper.type(
            df_to_type=df_the_data,
            dict_columns_to_type={
                "val 2": PandasDataframeTyper.str__type__date
            }
        )

        df_the_return: DataFrame = PandasDataframeSampler.compress(
            df_measures=df_the_data[[i_col for i_col in df_the_data.columns if i_col != "expected result"]],
            lst_key_columns=["key 1", "key 2"],
            lst_ordering_columns=["order 1", "order 2"],
            lst_value_columns=["val 1", "val 2"],
            b_keep_last=True,
        )

        assert_frame_equal(
            left=df_the_data[
                df_the_data["expected result"] == "filter in"
            ][df_the_return.columns].reset_index(drop=True),
            right=df_the_return.reset_index(drop=True),
            obj=f"Call 1",
            atol=0.00001,
            check_like=True  # To exclude index, and not to care about column ordering
        )

        self._logger.debug(f"Function '{stack()[0].filename} - {stack()[0].function}' is returning.")



import os
import unittest
from pathlib import Path
from logging import Logger, getLogger
from inspect import stack
from pandas import DataFrame, read_excel

from com_enovation.toolbox.data_handler.data_modifier.data_modification_feature_factory \
    import DataModificationFeatureFactory


class TestDataModificationFeatureFactory_AuditColumnStamped(unittest.TestCase):
    _logger: Logger = getLogger(__name__)

    def test_audit_column_stamped(self):
        self._logger.debug(f"Function '{stack()[0].filename} - {stack()[0].function}' is called")

        # We load the data extract
        df_data_extract: DataFrame = read_excel(
            Path(os.path.join(os.path.dirname(__file__),
                              '01.data_extract_for_audit_column.xlsx'))
        )

        # 1. We audit the column, grouping on one single column
        df_data_extract = DataModificationFeatureFactory(). \
            get_data_modification_feature__audit_column__stamped(
            s_column_to_audit="Jsg - Audited Column",
            s_column_to_stamp="Jsg - Stamp Column",
            s_column_audit="Jsg - Audited Column_audit",
            lst_key_columns=["Jsg - ID", "a column which does not exist"]
        )(df_data_extract)

        self.assertListEqual(
            list1=df_data_extract["Jsg - Audited Column_audit"].values.tolist(),
            list2=[
                1,
                1,
                3,
                4,
                4,
                6,
                7,
                8,
                8,
                10,
                11,
                11,
                13,
                14,
                15,

                16
            ],
            msg="1. We audit the column, grouping on one single column"
        )

        # 2. We audit the column, grouping on two columns
        df_data_extract = DataModificationFeatureFactory(). \
            get_data_modification_feature__audit_column__stamped(
            s_column_to_audit="Jsg - Audited Column",
            s_column_to_stamp="Jsg - Stamp Column",
            s_column_audit="Jsg - Audited Column_audit",
            lst_key_columns=["Jsg - ID", "Jsg - Snapshot"]
        )(df_data_extract)

        self.assertListEqual(
            list1=df_data_extract["Jsg - Audited Column_audit"].values.tolist(),
            list2=[
                1,
                1,
                3,
                4,
                4,
                6,
                7,
                8,
                9,
                10,
                11,
                11,
                13,
                14,
                15,
                16,
            ],
            msg="2. We audit the column, grouping on two columns"
        )

        # 3. We audit a column, using this same column as the stamp (equivalent to add the last audited value)
        df_data_extract = DataModificationFeatureFactory(). \
            get_data_modification_feature__audit_column__stamped(
            s_column_to_audit="Jsg - Audited Column",
            s_column_to_stamp="Jsg - Audited Column",
            s_column_audit="Jsg - Audited Column_audit",
            lst_key_columns=["Jsg - ID"]
        )(df_data_extract)

        self.assertListEqual(
            list1=df_data_extract["Jsg - Audited Column_audit"].values.tolist(),
            list2=[
                "A",
                "A",
                "B",
                "B",
                "B",
                "C",
                "A",
                "B",
                "B",
                "B",
                "C",
                "C",
                "A",
                "C",
                "B",
                "A"
            ],
            msg="We audit a column, using this same column as the stamp (equivalent to add the last audited value)"
        )

        # 5. Exception if we audit a column which does not exist
        with self.assertRaisesRegex(
                Exception,
                f"The column to audit 'Jsg - not existing column' does not exist in the dataframe, which contains the "
                f"following columns: 'Jsg - ID, Jsg - Snapshot, Jsg - Audited Column, Jsg - Stamp Column'."
        ):
            DataModificationFeatureFactory(). \
                get_data_modification_feature__audit_column__stamped(
                s_column_to_audit="Jsg - not existing column",
                s_column_to_stamp="Jsg - Stamp Column",
                s_column_audit="Jsg - Audited Column_audit",
                lst_key_columns=["Jsg - ID"]
            )(df_data_extract[["Jsg - ID", "Jsg - Snapshot", "Jsg - Audited Column", "Jsg - Stamp Column"]])

        # 6. Exception if we rely on a stamp column which does not exist
        with self.assertRaisesRegex(
                Exception,
                f"The column to stamp 'Jsg - not existing column' does not exist in the dataframe, which contains the "
                f"following columns: 'Jsg - ID, Jsg - Snapshot, Jsg - Audited Column, Jsg - Stamp Column'."
        ):
            DataModificationFeatureFactory(). \
                get_data_modification_feature__audit_column__stamped(
                s_column_to_audit="Jsg - Audited Column",
                s_column_to_stamp="Jsg - not existing column",
                s_column_audit="Jsg - Audited Column_audit",
                lst_key_columns=["Jsg - ID", "a column which does not exist"]
            )(df_data_extract[["Jsg - ID", "Jsg - Snapshot", "Jsg - Audited Column", "Jsg - Stamp Column"]])

        self._logger.debug(f"Function '{stack()[0].filename} - {stack()[0].function}' is returning")



import os
import re
import unittest
from datetime import datetime
from inspect import stack
from logging import Logger, getLogger
from pathlib import Path

from numpy import nan
from pandas import Series, DataFrame, read_excel
from pandas.testing import assert_series_equal

from com_enovation.toolbox.data_handler.data_modifier.data_modification_feature_factory import \
    DataModificationFeatureFactory


class TestEnovAddEvalLambda(unittest.TestCase):
    _logger: Logger = getLogger(__name__)

    def test_add_eval_column__lambda(self):
        self._logger.debug(f"Function '{stack()[0].filename} - {stack()[0].function}' is called")

        # We load the data extract
        df_data_extract: DataFrame = read_excel(
            Path(os.path.join(os.path.dirname(__file__), '01.data_extract.xlsx'))
        )

        df_the_modified_dataframe: DataFrame = DataModificationFeatureFactory(). \
            get_data_modification_feature__add_eval_columns(
            str_expressions=f"LastCommentAsTuple = @enov__lambda__process_comments__get_last_comment_as_tuple("
                            f"lst_s_people=[ProjectManager, Consultants], s_comments=Comments)\n"
                            f"LastCommentDate = @enov_get_record_at_index_from_tuple(LastCommentAsTuple, 0)\n"
                            f"LastCommentWriter = @enov_get_record_at_index_from_tuple(LastCommentAsTuple, 1)",
            lst_dependencies=[
                "tests.com.enovation.toolbox.data_handler.data_modifier."
                "test_get_data_modification_feature__add_eval_column."
                "test_data_modification_feature__add_eval_column__lambdas."
                "test_lambdas."
                "enov__lambda__process_comments__get_last_comment_as_tuple"],
            b_copy=True
        )(df_data_extract)

        assert_series_equal(
            left=df_data_extract["ExpectedResultDate"],
            right=df_the_modified_dataframe["LastCommentDate"],
            obj="Actual result DATE is not equal to expected result when calling lambda function process_comments",
            check_names=False
        )

        assert_series_equal(
            left=df_data_extract["ExpectedResultWriter"].replace(nan, None),
            right=df_the_modified_dataframe["LastCommentWriter"],
            obj="Actual result WRITER is not equal to expected result when calling lambda function process_comments",
            check_names=False
        )

        self._logger.debug(f"Function '{stack()[0].filename} - {stack()[0].function}' is returning")


def enov__lambda__process_comments__get_last_comment_as_tuple(
        s_comments: Series,
        lst_s_people: list[Series] = None,
        s_snapshot_date: Series = None
) -> Series:

    _s_people: None | Series = None
    _df_the_dataframe: None | DataFrame = None

    if lst_s_people is not None:

        if len(lst_s_people) > 0:
            _s_people: Series = lst_s_people[0].astype(str)

            if len(lst_s_people) > 1:
                for i_s_people in lst_s_people[1:]:
                    _s_people = _s_people+i_s_people.astype(str)

            _df_the_dataframe = DataFrame({
                "people": _s_people,
                "comments": s_comments
            })

    else:
        _df_the_dataframe = DataFrame({
            "comments": s_comments
        })

    if s_snapshot_date is not None:
        _df_the_dataframe = _df_the_dataframe.assign(snapshot_date=s_snapshot_date)

    def _get_last_comment_as_tuple(s_row: Series) -> tuple[datetime | None, str | None, str | None] | None:
        _the_return = None

        # We split the comments, comment by comment...
        _lst_comments_as_tuples: list[tuple[str, str]] = re.findall(
            string=s_row["comments"],
            pattern="^((?:(?:0[1-9])|(?:[1-2][0-9])|(?:3[0-1]))\\-"
                    "(?:Jan|Feb|Mar|Apr|May|Jun|Jul|Aug|Sep|Oct|Nov|Dec)\\-20\\d{2} "
                    "(?:\\d{2}\\:\\d{2}\\:\\d{2}) (?:AM|PM)) by ([^\\:]*) [\\:] (.*)",
            flags=re.MULTILINE
        )

        if len(_lst_comments_as_tuples) > 0:
            _dt_the_last_date: datetime | None = None
            _s_the_writer: str | None = None
            _s_the_msg: str | None = None

            for i_commment_as_tuple in _lst_comments_as_tuples:
                _dt_the_date: datetime = datetime.strptime(i_commment_as_tuple[0], "%d-%b-%Y %I:%M:%S %p")

                _b_in: bool = True

                if "snapshot_date" in s_row:
                    if _dt_the_date >= s_row["snapshot_date"]:
                        _b_in = False

                if "people" in s_row:
                    if i_commment_as_tuple[1] not in s_row["people"]:
                        _b_in = False

                if _b_in:

                    if _dt_the_last_date is None:
                        _dt_the_last_date = _dt_the_date
                        _s_the_writer = i_commment_as_tuple[1]
                        _s_the_msg = i_commment_as_tuple[2]

                    elif _dt_the_date > _dt_the_last_date:
                        _dt_the_last_date = _dt_the_date
                        _s_the_writer = i_commment_as_tuple[1]
                        _s_the_msg = i_commment_as_tuple[2]

            return _dt_the_last_date, _s_the_writer, _s_the_msg

        return None

    _df_the_dataframe["enov_last_comment"] = _df_the_dataframe.apply(
        func=_get_last_comment_as_tuple,
        axis=1,
    )

    return _df_the_dataframe["enov_last_comment"]



from inspect import stack
from logging import getLogger, Logger
from unittest import TestCase
from com_enovation.toolbox.excel_dashboard.widgets.widget import Widget


class TestWidget_FunctionProcessConditionalFormatting(TestCase):

    _logger: Logger = getLogger(__name__)

    _str_reference: str = "B3"
    _str_column_range_limit: str = "A:C"
    _str_row_range_limit = "2:5"

    _test_data__a_process_conditional_formatting__range: list = [
        {
            "label": "TC01 - Vanilla test case.",
            "str_range": "A@2:C@5",
            "result": "A4:C7"
        },
        {
            "label": "TC02 - Vanilla test case.",
            "str_range": "@A@1:B10",
            "result": "B3:B10"
        },
        {
            "label": "TC03 - Column range.",
            "str_range": "A:@B",
            "result": "A4:C7"
        },
        {
            "label": "TC04 - Row range.",
            "str_range": "1:@5",
            "result": "B1:D7"
        },
        {
            "label": "TC51 - dollars are forbidden.",
            "str_range": "$A5:B10",
            "exception": "There is '\\$' in the parameter str_range='\\$A5:B10', which is not expected\\."
        },
    ]

    def test_a_process_conditional_formatting__range(self):
        self._logger.debug(f"Function '{stack()[0].filename} - {stack()[0].function}' is called")

        for i_test_case in self._test_data__a_process_conditional_formatting__range:

            # if test case expected NOT to raise an exception
            if "result" in i_test_case:
                self.assertEqual(
                    first=i_test_case["result"],
                    second=Widget._process_conditional_formatting__range(
                        str_range=i_test_case["str_range"],
                        str_reference_address=self._str_reference,
                        str_column_range_limit=self._str_column_range_limit,
                        str_row_range_limit=self._str_row_range_limit
                    ),
                    msg=i_test_case["label"]
                )

            # Else, test case expected to raise an exception
            else:

                with self.assertRaisesRegex(
                        Exception,
                        i_test_case["exception"]
                ):
                    Widget._process_conditional_formatting__range(
                        str_range=i_test_case["str_range"],
                        str_reference_address=self._str_reference,
                        str_column_range_limit=self._str_column_range_limit,
                        str_row_range_limit=self._str_row_range_limit
                    )

        self._logger.debug(f"Function '{stack()[0].filename} - {stack()[0].function}' is returning")

    def test_a_process_conditional_formatting__multi_range(self):
        self._logger.debug(f"Function '{stack()[0].filename} - {stack()[0].function}' is called")

        _lst_correct_test_case__input: list[str] = [
            i_test_case["str_range"]
            for i_test_case in self._test_data__a_process_conditional_formatting__range
            if "result" in i_test_case
        ]

        _lst_correct_test_case__result: list[str] = [
            i_test_case["result"]
            for i_test_case in self._test_data__a_process_conditional_formatting__range
            if "result" in i_test_case
        ]

        _lst_incorrect_test_case__input: list[str] = [
            i_test_case["str_range"]
            for i_test_case in self._test_data__a_process_conditional_formatting__range
            if "exception" in i_test_case
        ]

        # We first test the successful concatenation
        self.assertEqual(
            first=" ".join(_lst_correct_test_case__result),
            second=Widget._process_conditional_formatting__multi_range(
                str_multi_range=" ".join(_lst_correct_test_case__input),
                str_reference_address=self._str_reference,
                str_column_range_limit=self._str_column_range_limit,
                str_row_range_limit=self._str_row_range_limit
            ),
            msg=f"Successful..."
        )

        # We go through each and every incorrect test case
        for i_test_case in _lst_incorrect_test_case__input:

            with self.assertRaisesRegex(
                    Exception,
                    ""
            ):
                Widget._process_conditional_formatting__multi_range(
                    str_multi_range=" ".join(_lst_correct_test_case__input+[i_test_case]),
                    str_reference_address=self._str_reference,
                    str_column_range_limit=self._str_column_range_limit,
                    str_row_range_limit=self._str_row_range_limit
                )

        self._logger.debug(f"Function '{stack()[0].filename} - {stack()[0].function}' is returning")



from datetime import datetime
from inspect import stack
from logging import Logger, getLogger
from unittest import TestCase

from jsonschema import ValidationError

from com_enovation.toolbox.excel_dashboard.config_checker import ConfigChecker


class TestWorksheetConditionalFormat(TestCase):

    _logger: Logger = getLogger(__name__)

    _dict_a_full_format_cell_equal: dict = {
        # most common conditional formatting type. It is used when a format is applied to a cell based on a
        # simple criterion
        "type": "cell",

        # to set the criteria by which the cell data will be evaluated. It has no default value. The most
        # common criteria as applied to {'type': 'cell'} are:
        # - between
        # - not between
        # - equal to
        # - not equal to
        # - greater than
        # - less than
        # - greater than or equal to
        # - less than or equal to
        "criteria": "equal to",

        # generally used along with the criteria parameter to set the rule by which the cell data will be
        # evaluated. If the type is cell and the value is a string then it should be double quoted, as
        # required by Excel
        "value": '"Failed"',

        # to specify the format that will be applied to the cell when the conditional formatting criterion is met
        "format": {'bold': True, 'font_color': 'red'},

        # to set the “stop if true” feature of a conditional formatting rule when more than one rule is
        # applied to a cell or a range of cells. When this parameter is set then subsequent rules are not
        # evaluated if the current rule is true:
        "stop_if_true": True,

        # The multi_range option is used to extend a conditional format over non-contiguous ranges.
        # It is possible to apply the conditional format to different cell ranges in a worksheet using multiple
        # calls to conditional_format(). However, as a minor optimization it is also possible in Excel to apply
        # the same conditional format to different non-contiguous cell ranges.
        # This is replicated in conditional_format() using the multi_range option. The range must contain the
        # primary range for the conditional format and any others separated by spaces.
        "multi_range": "A1:A2 B3:B4 D15"
    }

    _dict_a_full_format_cell_between: dict = {
        # most common conditional formatting type. It is used when a format is applied to a cell based on a
        # simple criterion
        "type": "cell",

        # to set the criteria by which the cell data will be evaluated. It has no default value. The most
        # common criteria as applied to {'type': 'cell'} are:
        # - between
        # - not between
        # - equal to
        # - not equal to
        # - greater than
        # - less than
        # - greater than or equal to
        # - less than or equal to
        "criteria": "between",

        # generally used along with the criteria parameter to set the rule by which the cell data will be
        # evaluated. If the type is cell and the value is a string then it should be double quoted, as
        # required by Excel
        # "value": '"Failed"',

        # to set the lower limiting value when the criteria is either 'between' or 'not between'
        "minimum": 2,

        # to set the upper limiting value when the criteria is either 'between' or 'not between'
        "maximum": 5,

        # to specify the format that will be applied to the cell when the conditional formatting criterion is met
        "format": {'bold': True, 'font_color': 'red'},

        # to set the “stop if true” feature of a conditional formatting rule when more than one rule is
        # applied to a cell or a range of cells. When this parameter is set then subsequent rules are not
        # evaluated if the current rule is true:
        "stop_if_true": True,

        # The multi_range option is used to extend a conditional format over non-contiguous ranges.
        # It is possible to apply the conditional format to different cell ranges in a worksheet using multiple
        # calls to conditional_format(). However, as a minor optimization it is also possible in Excel to apply
        # the same conditional format to different non-contiguous cell ranges.
        # This is replicated in conditional_format() using the multi_range option. The range must contain the
        # primary range for the conditional format and any others separated by spaces.
        "multi_range": "A1:A2 B3:B4"
    }

    _dict_a_full_format_date_equal: dict = {
        # date type is similar the cell type and uses the same criteria and values. However, the value, minimum
        # and maximum properties are specified as a datetime object
        "type": "date",

        # criteria is the same as {'type': 'date'}
        "criteria": "equal to",

        # the value, minimum and maximum properties are specified as a datetime object
        "value": f"{datetime.strptime('2011-01-01', '%Y-%m-%d')}",

        # format, stop_if_true and multi_range are the same as {'type': 'date'}
        "format": {'bold': True, 'font_color': 'red'},
        "stop_if_true": True,
        "multi_range": "A1:A2 B3:B4"
    }

    _dict_a_full_format_date_between: dict = {
        # date type is similar the cell type and uses the same criteria and values. However, the value, minimum
        # and maximum properties are specified as a datetime object
        "type": "date",

        # criteria is the same as {'type': 'date'}
        "criteria": "between",

        # the value, minimum and maximum properties are specified as a datetime object
        "minimum": f"{datetime.strptime('2011-01-01', '%Y-%m-%d')}",
        "maximum": f"{datetime.strptime('2011-01-01', '%Y-%m-%d')}",

        # format, stop_if_true and multi_range are the same as {'type': 'date'}
        "format": {'bold': True, 'font_color': 'red'},
        "stop_if_true": True,
        "multi_range": "A1:A2 B3:B4"
    }

    _dict_a_full_format_text: dict = {
        # to specify Excel’s “Specific Text” style conditional format. It is used to do simple string matching
        # using the criteria and value parameters:
        "type": "text",

        # the criteria can have one of the following values:
        # - containing
        # - not containing
        # - begins with
        # - ends with
        "criteria": "containing",

        # The value parameter should be a string or single character
        "value": "jsg",

        # format, stop_if_true and multi_range are the same as {'type': 'date'}
        "format": {'bold': True, 'font_color': 'red'},
        "stop_if_true": True,
        "multi_range": "A1:A2 B3:B4"
    }

    _dict_a_full_format_formula: dict = {
        # to specify a conditional format based on a user defined formula
        "type": "formula",

        # formula is specified in the criteria, written with the US style separator/range operator which is a
        # comma (not semi-colon) and should follow the same rules as write_formula().
        # See Non US Excel functions and syntax for a full explanation:
        # https://xlsxwriter.readthedocs.io/working_with_formulas.html
        "criteria": "=today()",

        # format, stop_if_true and multi_range are the same as {'type': 'date'}
        "format": {'bold': True, 'font_color': 'red'},
        "stop_if_true": True,
        "multi_range": "A1:A2 B3:B4"
    }

    def test_111_cell_equal_successful(self):
        self._logger.debug(f"Function '{stack()[0].filename} - {stack()[0].function}' is called")

        _the_config_checker = ConfigChecker(
            str_base_json_schema_id="https://enovation.com/xlsxwriter/worksheet/conditional_format/options",
            lst_widgets_json_schemas=[]
        )

        _the_config_checker.validate(
            dict_to_validate=self._dict_a_full_format_cell_equal
        )

        self._logger.debug(f"Function '{stack()[0].filename} - {stack()[0].function}' is returning")

    def test_112_cell_equal_unexpected_tag(self):
        self._logger.debug(f"Function '{stack()[0].filename} - {stack()[0].function}' is called")

        _the_config_checker = ConfigChecker(
            str_base_json_schema_id="https://enovation.com/xlsxwriter/worksheet/conditional_format/options",
            lst_widgets_json_schemas=[]
        )

        _the_altered_dict: dict = dict(self._dict_a_full_format_cell_equal)
        _the_altered_dict["jsg"] = "bouh"

        with self.assertRaisesRegex(
                ValidationError,
                "{'type': 'cell', 'criteria': 'equal to', 'value': '\"Failed\"', 'format': {'bold': True, "
                "'font_color': 'red'}, 'stop_if_true': True, 'multi_range': 'A1:A2 B3:B4 D15', 'jsg': 'bouh'} is not valid "
                "under any of the given schemas"
        ):
            _the_config_checker.validate(
                dict_to_validate=_the_altered_dict
            )

        self._logger.debug(f"Function '{stack()[0].filename} - {stack()[0].function}' is returning")

    def test_121_cell_between_successful(self):
        self._logger.debug(f"Function '{stack()[0].filename} - {stack()[0].function}' is called")

        _the_config_checker = ConfigChecker(
            str_base_json_schema_id="https://enovation.com/xlsxwriter/worksheet/conditional_format/options",
            lst_widgets_json_schemas=[]
        )

        _the_config_checker.validate(
            dict_to_validate=self._dict_a_full_format_cell_between
        )

        self._logger.debug(f"Function '{stack()[0].filename} - {stack()[0].function}' is returning")

    def test_122_cell_between_unexpected_tag(self):
        self._logger.debug(f"Function '{stack()[0].filename} - {stack()[0].function}' is called")

        _the_config_checker = ConfigChecker(
            str_base_json_schema_id="https://enovation.com/xlsxwriter/worksheet/conditional_format/options",
            lst_widgets_json_schemas=[]
        )

        _the_altered_dict: dict = dict(self._dict_a_full_format_cell_between)
        _the_altered_dict["jsg"] = "bouh"

        with self.assertRaisesRegex(
                ValidationError,
                "{'type': 'cell', 'criteria': 'between', 'minimum': 2, 'maximum': 5, 'format': {'bold': True, "
                "'font_color': 'red'}, 'stop_if_true': True, 'multi_range': 'A1:A2 B3:B4', 'jsg': 'bouh'} is not valid "
                "under any of the given schemas"
        ):
            _the_config_checker.validate(
                dict_to_validate=_the_altered_dict
            )

        self._logger.debug(f"Function '{stack()[0].filename} - {stack()[0].function}' is returning")

    def test_211_date_equal_successful(self):
        self._logger.debug(f"Function '{stack()[0].filename} - {stack()[0].function}' is called")

        _the_config_checker = ConfigChecker(
            str_base_json_schema_id="https://enovation.com/xlsxwriter/worksheet/conditional_format/options",
            lst_widgets_json_schemas=[]
        )

        _the_config_checker.validate(
            dict_to_validate=self._dict_a_full_format_date_equal
        )

        self._logger.debug(f"Function '{stack()[0].filename} - {stack()[0].function}' is returning")

    def test_212_date_equal_unexpected_tag(self):
        self._logger.debug(f"Function '{stack()[0].filename} - {stack()[0].function}' is called")

        _the_config_checker = ConfigChecker(
            str_base_json_schema_id="https://enovation.com/xlsxwriter/worksheet/conditional_format/options",
            lst_widgets_json_schemas=[]
        )

        _the_altered_dict: dict = dict(self._dict_a_full_format_date_equal)
        _the_altered_dict["jsg"] = "bouh"

        with self.assertRaisesRegex(
                ValidationError,
                "{'type': 'date', 'criteria': 'equal to', 'value': '2011-01-01 00:00:00', 'format': {'bold': True, "
                "'font_color': 'red'}, 'stop_if_true': True, 'multi_range': 'A1:A2 B3:B4', 'jsg': 'bouh'} is not valid "
                "under any of the given schemas"
        ):
            _the_config_checker.validate(
                dict_to_validate=_the_altered_dict
            )

        self._logger.debug(f"Function '{stack()[0].filename} - {stack()[0].function}' is returning")

    def test_221_date_between_successful(self):
        self._logger.debug(f"Function '{stack()[0].filename} - {stack()[0].function}' is called")

        _the_config_checker = ConfigChecker(
            str_base_json_schema_id="https://enovation.com/xlsxwriter/worksheet/conditional_format/options",
            lst_widgets_json_schemas=[]
        )

        _the_config_checker.validate(
            dict_to_validate=self._dict_a_full_format_date_between
        )

        self._logger.debug(f"Function '{stack()[0].filename} - {stack()[0].function}' is returning")

    def test_222_date_between_unexpected_tag(self):
        self._logger.debug(f"Function '{stack()[0].filename} - {stack()[0].function}' is called")

        _the_config_checker = ConfigChecker(
            str_base_json_schema_id="https://enovation.com/xlsxwriter/worksheet/conditional_format/options",
            lst_widgets_json_schemas=[]
        )

        _the_altered_dict: dict = dict(self._dict_a_full_format_date_between)
        _the_altered_dict["jsg"] = "bouh"

        with self.assertRaisesRegex(
                ValidationError,
                "{'type': 'date', 'criteria': 'between', 'minimum': '2011-01-01 00:00:00', 'maximum': "
                "'2011-01-01 00:00:00', 'format': {'bold': True, 'font_color': 'red'}, 'stop_if_true': True, "
                "'multi_range': 'A1:A2 B3:B4', 'jsg': 'bouh'} is not valid under any of the given schemas"
        ):
            _the_config_checker.validate(
                dict_to_validate=_the_altered_dict
            )

        self._logger.debug(f"Function '{stack()[0].filename} - {stack()[0].function}' is returning")

    def test_311_formula_successful(self):
        self._logger.debug(f"Function '{stack()[0].filename} - {stack()[0].function}' is called")

        _the_config_checker = ConfigChecker(
            str_base_json_schema_id="https://enovation.com/xlsxwriter/worksheet/conditional_format/options",
            lst_widgets_json_schemas=[]
        )

        _the_config_checker.validate(
            dict_to_validate=self._dict_a_full_format_formula
        )

        self._logger.debug(f"Function '{stack()[0].filename} - {stack()[0].function}' is returning")

    def test_312_formula_unexpected_tag(self):
        self._logger.debug(f"Function '{stack()[0].filename} - {stack()[0].function}' is called")

        _the_config_checker = ConfigChecker(
            str_base_json_schema_id="https://enovation.com/xlsxwriter/worksheet/conditional_format/options",
            lst_widgets_json_schemas=[]
        )

        _the_altered_dict: dict = dict(self._dict_a_full_format_formula)
        _the_altered_dict["jsg"] = "bouh"

        with self.assertRaisesRegex(
                ValidationError,
                "{'type': 'formula', 'criteria': '=today\\(\\)', 'format': {'bold': True, 'font_color': 'red'}, "
                "'stop_if_true': True, 'multi_range': 'A1:A2 B3:B4', 'jsg': 'bouh'} is not valid under any of the "
                "given schemas"
        ):
            _the_config_checker.validate(
                dict_to_validate=_the_altered_dict
            )

        self._logger.debug(f"Function '{stack()[0].filename} - {stack()[0].function}' is returning")

    def test_411_text_successful(self):
        self._logger.debug(f"Function '{stack()[0].filename} - {stack()[0].function}' is called")

        _the_config_checker = ConfigChecker(
            str_base_json_schema_id="https://enovation.com/xlsxwriter/worksheet/conditional_format/options",
            lst_widgets_json_schemas=[]
        )

        _the_config_checker.validate(
            dict_to_validate=self._dict_a_full_format_text
        )

        self._logger.debug(f"Function '{stack()[0].filename} - {stack()[0].function}' is returning")

    def test_412_text_unexpected_tag(self):
        self._logger.debug(f"Function '{stack()[0].filename} - {stack()[0].function}' is called")

        _the_config_checker = ConfigChecker(
            str_base_json_schema_id="https://enovation.com/xlsxwriter/worksheet/conditional_format/options",
            lst_widgets_json_schemas=[]
        )

        _the_altered_dict: dict = dict(self._dict_a_full_format_text)
        _the_altered_dict["jsg"] = "bouh"

        with self.assertRaisesRegex(
                ValidationError,
                "{'type': 'text', 'criteria': 'containing', 'value': 'jsg', 'format': {'bold': True, 'font_color': "
                "'red'}, 'stop_if_true': True, 'multi_range': 'A1:A2 B3:B4', 'jsg': 'bouh'} is not valid under any of "
                "the given schemas"
        ):
            _the_config_checker.validate(
                dict_to_validate=_the_altered_dict
            )

        self._logger.debug(f"Function '{stack()[0].filename} - {stack()[0].function}' is returning")

    def test_511_empty(self):
        self._logger.debug(f"Function '{stack()[0].filename} - {stack()[0].function}' is called")

        _the_config_checker = ConfigChecker(
            str_base_json_schema_id="https://enovation.com/xlsxwriter/worksheet/conditional_format/options",
            lst_widgets_json_schemas=[]
        )

        with self.assertRaisesRegex(
                ValidationError,
                "{} does not have enough properties"
        ):
            _the_config_checker.validate(
                dict_to_validate={}
            )

        self._logger.debug(f"Function '{stack()[0].filename} - {stack()[0].function}' is returning")

    def test_04_incorrect_format(self):
        self._logger.debug(f"Function '{stack()[0].filename} - {stack()[0].function}' is called")

        _the_config_checker = ConfigChecker(
            str_base_json_schema_id="https://enovation.com/xlsxwriter/worksheet/conditional_format/options",
            lst_widgets_json_schemas=[]
        )

        _the_altered_dict: dict = dict(self._dict_a_full_format_cell_equal)
        _the_altered_dict["format"] = {
            "font_name": "Times New Roman",
            "font_size": "15",
        }

        with self.assertRaisesRegex(
                ValidationError,
                f"'15' is not of type 'integer'"
        ):
            _the_config_checker.validate(
                dict_to_validate=_the_altered_dict
            )

        self._logger.debug(f"Function '{stack()[0].filename} - {stack()[0].function}' is returning")



from inspect import stack
from logging import Logger, getLogger
from unittest import TestCase

from jsonschema import ValidationError

from com_enovation.toolbox.excel_dashboard.config_checker import ConfigChecker


class TestColors(TestCase):

    _logger: Logger = getLogger(__name__)

    def test_01_validated(self):
        self._logger.debug(f"Function '{stack()[0].filename} - {stack()[0].function}' is called")

        _the_config_checker = ConfigChecker(
            str_base_json_schema_id="https://enovation.com/xlsxwriter/colors",
            lst_widgets_json_schemas=[]
        )

        for i_color in [
            "red",
            "black",
            "#012345",
            "#A1D2B1"
        ]:
            # noinspection PyTypeChecker
            _the_config_checker.validate(dict_to_validate=i_color)

        self._logger.debug(f"Function '{stack()[0].filename} - {stack()[0].function}' is returning")

    def test_02_not_validated(self):
        self._logger.debug(f"Function '{stack()[0].filename} - {stack()[0].function}' is called")

        _the_config_checker = ConfigChecker(
            str_base_json_schema_id="https://enovation.com/xlsxwriter/colors",
            lst_widgets_json_schemas=[]
        )

        for i_color in [
            "redx",
            "#01234G"
        ]:
            with self.assertRaisesRegex(
                    ValidationError,
                    f"is not valid under any of the given schemas"
            ):
                # noinspection PyTypeChecker
                _the_config_checker.validate(dict_to_validate=i_color)

        self._logger.debug(f"Function '{stack()[0].filename} - {stack()[0].function}' is returning")



from inspect import stack
from unittest import TestCase
from logging import Logger, getLogger


from com_enovation.helper.pandas_dataframe_typer import PandasDataframeTyper
from com_enovation.toolbox.backlog.bck_compare.bck_comparer import BacklogComparer


class TestFunctionProcessDictColumns(TestCase):
    _logger: Logger = getLogger(__name__)

    def test_001(self):
        self._logger.debug(f"Function '{stack()[0].filename} - {stack()[0].function}' is called")

        # From the dict_columns, we prepare the various sub-dictionaries/ lists
        dict_the_columns_types: dict[str]   # A dictionary to properly type the dataframes (both key and comparable col)
        lst_the_compared_columns: list[str] # A list of the columns' labels to compare
        lst_the_key_columns: list[str]      # A list of the key columns' labels to merge dataframes
        dict_the_columns_comparison_logics: dict    # A dictionary to properly associate comparison functions to columns

        # First, we process the dict_columns, and raise an exception should its format be incorrect
        dict_columns_types, lst_compared_columns, lst_key_columns, dict_the_columns_comparison_logics = \
            BacklogComparer()._process_dict_columns(
                dict_columns={
                    BacklogComparer.str__dict_columns__json_tag__key_columns: {
                        "key1": {
                            BacklogComparer.str__dict_columns__json_tag__type: PandasDataframeTyper.str__type__str,
                        }
                    },
                    BacklogComparer.str__dict_columns__json_tag__compared_columns: {
                        "col1": {
                            BacklogComparer.str__dict_columns__json_tag__type: PandasDataframeTyper.str__type__str,
                        }
                    }
                }
            )

        self.assertDictEqual(
            d1={
                "key1": PandasDataframeTyper.str__type__str,
                "col1": PandasDataframeTyper.str__type__str
            },
            d2=dict_columns_types,
            msg=f"dict_columns_types not equal. D1 as expected results"
        )

        self._logger.debug(f"Function '{stack()[0].filename} - {stack()[0].function}' is returning")



import logging
import os

import click
from click import Context


_logger: logging.Logger = logging.getLogger(__name__)


@click.command('dbg')
@click.pass_context
def debug(ctx_context: Context):
    """
    Command to run whenever you need to plug in a debugger on a given process. The command display the process id,
    so you can plug the debugger, and wait for you to confirm in the terminal to resume execution...
    """

    click.echo(f"Process ID: {os.getpid()}")

    click.confirm('Ready to continue?', abort=True, default=True)




import os
import unittest
from pathlib import Path
from logging import Logger, getLogger
from inspect import stack
from pandas import DataFrame, read_excel, Series, MultiIndex, Index
from pandas.testing import assert_series_equal

from com_enovation.toolbox.data_handler.data_checker.data_check_feature_factory import DataCheckFeatureFactory


class TestDataCheckFeatureFactory_CheckDuplicate(unittest.TestCase):
    _logger: Logger = getLogger(__name__)

    def test_check_duplicate__1_column(self):
        self._logger.debug(f"Function '{stack()[0].filename} - {stack()[0].function}' is called")

        # We load the data extract
        df_data_extract: DataFrame = \
            read_excel(Path(os.path.join(os.path.dirname(__file__), '01.data_extract.xlsx')))

        # Run 1: duplicates found, as data extract contains 3 snapshots, so most id will be duplicated
        s_the_return: Series = DataCheckFeatureFactory().get_data_check_feature__check_duplicate(
            columns="Jsg - ID"
        )(df_data_extract)

        assert_series_equal(
            left=s_the_return.sort_index(),
            right=Series(
                data=[3, 3, 3, 3, 3],
                index=Index(
                    data=["JSG1384", "JSG1371", "JSG1365", "JSG1302", "JSG1338"],
                    name="Jsg - ID"
                ),
                name=DataCheckFeatureFactory.const_fn_data_extract_checker__count
            ).sort_index(),
            obj="Check 1: most opp_id are duplicated. We expect 5 duplicates.",
            check_exact=True
        )

        # Run 2: no duplicate found, because if we only select one snapshot, then we do not expect any duplicate
        s_the_return: Series = DataCheckFeatureFactory().get_data_check_feature__check_duplicate(
            columns="Jsg - ID"
        )(df_data_extract[df_data_extract["Jsg - Snapshot"] == 1])

        self.assertIsNone(
            obj=s_the_return,
            msg="Check 2: no duplicate, we expect a null value!"
        )

        self._logger.debug(f"Function '{stack()[0].filename} - {stack()[0].function}' is returning")

    def test_check_duplicate__2_columns(self):
        self._logger.debug(f"Function '{stack()[0].filename} - {stack()[0].function}' is called")

        # We load the data extract
        df_data_extract: DataFrame = \
            read_excel(Path(os.path.join(os.path.dirname(__file__), '01.data_extract.xlsx')))

        # Run 1: no duplicate found, as no duplicate in the source file across id and snapshot
        s_the_return: Series = DataCheckFeatureFactory().get_data_check_feature__check_duplicate(
            columns=["Jsg - ID", "Jsg - Snapshot"],
        )(df_data_extract)

        self.assertIsNone(
            obj=s_the_return,
            msg="Check 1: no duplicate, we expect a null value!"
        )

        # Run 2: duplicates found after altering the data extract
        df_data_extract.at[1, "Jsg - Snapshot"] = 1
        df_data_extract.at[2, "Jsg - Snapshot"] = 1

        df_data_extract.at[3, "Jsg - Snapshot"] = 2
        df_data_extract.at[4, "Jsg - Snapshot"] = 2
        df_data_extract.at[5, "Jsg - Snapshot"] = 2
        df_data_extract.at[6, "Jsg - Snapshot"] = 2
        df_data_extract.at[7, "Jsg - Snapshot"] = 2

        s_the_return: Series = DataCheckFeatureFactory().get_data_check_feature__check_duplicate(
            columns=["Jsg - ID", "Jsg - Snapshot"],
        )(df_data_extract)

        assert_series_equal(
            left=s_the_return.sort_index(),
            right=Series(
                data=[3, 3, 2],
                index=MultiIndex.from_arrays(
                    [["JSG1384", "JSG1371", "JSG1365"], [1, 2, 2]],
                    names=("Jsg - ID", "Jsg - Snapshot")
                ),
                name=DataCheckFeatureFactory.const_fn_data_extract_checker__count
            ).sort_index(),
            obj="Check 2: duplicates, we expect one combination duplicated 3 times!",
            check_exact=True
        )

        self._logger.debug(f"Function '{stack()[0].filename} - {stack()[0].function}' is returning")

    def test_check_duplicate__exception(self):
        self._logger.debug(f"Function '{stack()[0].filename} - {stack()[0].function}' is called")

        # We load the data extract
        df_data_extract: DataFrame = \
            read_excel(Path(os.path.join(os.path.dirname(__file__), '01.data_extract.xlsx')))

        # Exception 1: single column does not exist in extract
        with self.assertRaisesRegex(
                Exception,
                f"Column 'Jsg - not defined' is missing from the data extract to check."
        ):
            DataCheckFeatureFactory().get_data_check_feature__check_duplicate(
                columns="Jsg - not defined"
            )(df_data_extract)

        # Exception 2: multiple column does not exist in extract
        with self.assertRaisesRegex(
                Exception,
                f"Column 'Jsg - not defined' is missing from the data extract to check."
        ):
            DataCheckFeatureFactory().get_data_check_feature__check_duplicate(
                columns=["Jsg - not defined", "Jsg - and a second one"]
            )(df_data_extract)

        self._logger.debug(f"Function '{stack()[0].filename} - {stack()[0].function}' is returning")
